from datetime import datetime
import re
from typing import Iterable, List, Set, Optional, Iterator, Union, Tuple, NamedTuple
from pathlib import Path


from ..common import PreVisit, get_logger, Extraction, Url, Loc, from_epoch, echain, extract_urls, PathIsh


import orgparse # type: ignore
from orgparse.date import gene_timestamp_regex, OrgDate # type: ignore
from orgparse.node import OrgNode # type: ignore


rgx = re.compile(gene_timestamp_regex(brtype='inactive'), re.VERBOSE)


# TODO should be include child note contents??
# e.g. what to do with
"""
* something .... http://reddit.org
** subnote
** subnote
"""

class Parsed(NamedTuple):
    dt: Optional[datetime]
    heading: str

def _parse_node(n: OrgNode) -> Parsed:
    if n.is_root():
        return Parsed(dt=None, heading='')

    heading = n.get_heading('raw')
    pp = n.properties or {}
    createds = pp.get('CREATED', None)
    if createds is None:
        # TODO replace with 'match', but need to strip off priority etc first?
        # see _parse_heading in orgparse
        m = rgx.search(heading)
        if m is not None:
            createds = m.group(0) # could be None
            # TODO a bit hacky..
            heading = heading.replace(createds + ' ', '')
    if createds is not None:
        [odt] = OrgDate.list_from_str(createds)
        dt = odt.start
    else:
        dt = None
    return Parsed(dt=dt, heading=heading)


def _get_heading(n: OrgNode):
    return '' if n.is_root() else n.get_heading(format='raw') # TODO convert links to html?


# TODO maybe line by line? or restrict length? not sure..
def _get_body(n: OrgNode):
    if n.is_root():
        return '\n'.join(n._lines)
    else:
        return n.get_body(format='raw')


def walk_node(*, node: OrgNode, dt: datetime) -> Iterator[Union[Tuple[Parsed, OrgNode], Exception]]:
    try:
        parsed = _parse_node(node)
    except Exception as e:
        yield e
    else:
        if parsed.dt is None:
            parsed = parsed._replace(dt=dt)
        else:
            dt = parsed.dt
    yield parsed, node

    for c in node.children:
        yield from walk_node(node=c, dt=dt)


def iter_urls(n: OrgNode) -> Iterator[Union[Url, Exception]]:
    logger = get_logger()
    try:
        heading = _get_heading(n)
    except Exception as e:
        logger.exception(e)
        yield e
    else:
        yield from extract_urls(heading, syntax='org')

    try:
        content = _get_body(n)
    except Exception as e:
        logger.exception(e)
        yield e
    else:
        yield from extract_urls(content, syntax='org')


def extract_from_file(fname: PathIsh) -> Iterator[Extraction]:
    """
    Note that org-mode doesn't keep timezone, so we don't really have choice but make it tz-agnostic
    """
    path = Path(fname)
    o = orgparse.loads(path.read_text())
    # meh. but maybe ok to start with?
    root = o.root

    fallback_dt = datetime.fromtimestamp(path.stat().st_mtime)

    ex = RuntimeError(f'while extracting from {fname}')

    for wr in walk_node(node=root, dt=fallback_dt):
        if isinstance(wr, Exception):
            yield echain(ex, wr)
            continue

        (pn, n) = wr
        dt = pn.dt
        assert dt is not None # shouldn't be because of fallback
        for r in iter_urls(n):
            try:
                # TODO get body recursively? not sure
                ctx = pn.heading + '\n' + _get_body(n)
            except Exception as e:
                yield echain(ex, e)
                ctx = 'ERROR' # TODO more context?

            if isinstance(r, Url):
                yield PreVisit(
                    url=r,
                    dt=dt,
                    locator=Loc.file(fname), # TODO line number
                    context=ctx,
                )
            else: # error
                yield echain(ex, r)



#!/usr/bin/env python3
import sys
from pathlib import Path
from subprocess import check_call


def convert(path: Path):
    suf = '.mp4'
    if path.suffix == suf:
        # makes it easier for shell globbing...
        path = path.with_suffix('')

    inp  = path.with_suffix(suf)
    assert inp.exists(), inp
    subs = path.with_suffix('.ssa')
    webm = path.with_suffix('.webm')


    # jeez... https://video.stackexchange.com/a/28276/29090
    # otherwise quiality sucks, e.g. letters are grainy
    passfile = path.with_suffix(".pass0")
    for stage in [
            f'-b:v 0  -crf 30  -pass 1 -passlogfile {passfile} -an -f webm /dev/null',
            f'-b:v 0  -crf 30  -pass 2 -passlogfile {passfile} {webm}' if all(
                x not in str(inp) for x in (
                    # fucking hell, it segfaults...
                    'child-visits-2',
                    'highlights',
                )) else str(webm),
    ]:
        check_call([
            'ffmpeg',
            # TODO display banner if running interactively??
            # '-hide_banner', '-loglevel', 'panic', # less spam
            '-y', # allow overwrite
            '-i', inp,
            '-vf', f"ass={subs}",
            *stage.split(),
        ]) # TODO cwd??


if __name__ == '__main__':
    paths = list(map(Path, sys.argv[1:]))
    from concurrent.futures import ThreadPoolExecutor
    with ThreadPoolExecutor() as pool:
        for _ in pool.map(convert, paths):
            # need to force the iterator
            pass



from promnesia.sources.org import extract_from_file

from common import tdata

def test_org_indexer():
    items = list(extract_from_file(tdata('auto/orgs/file.org')))
    assert len(items) == 3

    cpp = items[1]
    assert cpp.url == 'https://www.youtube.com/watch?v=rHIkrotSwcc'

    cozy = items[2]
    assert cozy.url == 'https://twitter.com/Mappletons/status/1255221220263563269'


def test_org_indexer_2():
    items = list(extract_from_file(tdata('auto/orgs/file3.org')))

    assert len(items) == 6
    for i in items:
        assert i.dt.tzinfo is None, i
    assert items[0].url == 'https://www.reddit.com/r/androidapps/comments/4i36z9/how_you_use_your_android_to_the_maximum/d2uq24i'
    assert items[1].url == 'https://link.com'
    assert items[-2].url == 'https://en.wikipedia.org/wiki/Resilio_Sync'
    # TODO shit def need org specific url extractor (and then extract from everything remaining)
    # assert results[-1].url == 'https://en.wikipedia.org/wiki/InterPlanetary_File_System'


def test_heading():
    items = list(extract_from_file(tdata('auto/orgs/file2.org')))
    assert {i.url for i in items} == {
        'https://en.wikipedia.org/wiki/Computational_topology',
        'http://graphics.stanford.edu/courses/cs468-09-fall/',
        'https://en.wikipedia.org/wiki/Triangulation_(topology)',
        'https://en.wikipedia.org/wiki/Digital_manifold',
    }





#!/usr/bin/env python3
import argparse

import os
from pathlib import Path
from subprocess import check_call, run

from .. import root
from .. import server

SYSTEMD_CONFIG = """
[Unit]
Description=Promnesia browser extension backend

[Install]
WantedBy=default.target

[Service]
ExecStart={launcher} {extra_args}
Type=simple
Restart=always
"""


def systemd(*args, method=check_call):
    method([
        'systemctl', '--no-pager', '--user', *args,
    ])


def install(args):
    unit_name = args.unit_name
    out = Path(f'~/.config/systemd/user/{unit_name}').expanduser()
    print(f"Writing systemd config to {out}:")

    # ugh. we want to know whether we're invoked 'properly' as an executable or ad-hoc via scripts/promnesia
    if os.environ.get('DIRTY_RUN') is not None:
        launcher = str(root() / 'scripts/promnesia')
    else:
        # must be installed, so available in PATH
        import distutils.spawn
        exe = distutils.spawn.find_executable('promnesia'); assert exe is not None
        launcher = exe # older systemd wants absolute paths..

    extra_args = f'serve --db {args.db} --timezone {args.timezone} --port {args.port}'

    out.parent.mkdir(parents=True, exist_ok=True) # sometimes systemd dir doesn't exist
    out.write_text(SYSTEMD_CONFIG.format(
        launcher=launcher,
        extra_args=extra_args,
    ))

    try:
        systemd('stop' , unit_name, method=run) # ignore errors here if it wasn't running in the first place
        systemd('daemon-reload')
        systemd('enable', unit_name)
        systemd('start' , unit_name)
        systemd('status', unit_name)
    except Exception as e:
        print(f"Something has gone wrong... you might want to use 'journalctl --user -u {unit_name}' to investigate")
        raise e


def setup_parser(p: argparse.ArgumentParser):
    p.add_argument('--unit-name', type=str, default='promnesia.service', help='Systemd unit name')
    server.setup_parser(p)



from typing import Iterable

from ..common import Extraction, Visit, Loc, PathIsh, extract_urls

import my.roamresearch as RR


def index() -> Iterable[Extraction]:
    roam = RR.roam()
    for node in roam.traverse():
        yield from _collect(node)


def _collect(node: RR.Node) -> Iterable[Extraction]:
    title = node.title
    body  = node.body or ''
    if title is None:
        # most notes don't have title, so we just take the first line instead..
        lines = body.splitlines(keepends=True)
        if len(lines) > 0:
            title = lines[0]
            body = ''.join(lines)
    title = title or ''

    full = title + '\n' + body

    urls = extract_urls(full)
    if len(urls) == 0:
        return

    loc = Loc.make(
        title=node.path,
        href=node.permalink,
    )
    for u in urls:
        yield Visit(
            url=u,
            dt=node.created,
            context=body,
            locator=loc,
        )



"""
Some combinators for jq
"""

def pipe(*queries):
    return ' | '.join(queries)

def jdel(q):
    return f'del({q})'

def jq_del_all(*keys, split_by=10):
    parts = []
    # TODO shit. looks like query might be too long for jq...
    for q in range(0, len(keys), split_by):
        kk = keys[q: q + split_by]
        parts.append(jdel('.. | ({})'.format(', '.join('.' + k + '?' for k in kk))))
    return pipe(*parts)



from collections.abc import Sized
from datetime import datetime, date
import os.path
import re
from typing import NamedTuple, Set, Iterable, Dict, TypeVar, Callable, List, Optional, Union, Any, Collection, Sequence, Tuple, TypeVar
from pathlib import Path
from glob import glob
import itertools
import logging
from functools import lru_cache
import traceback
import pytz
import warnings

from .cannon import CanonifyException, canonify


T = TypeVar('T')
Res = Union[T, Exception]

PathIsh = Union[str, Path]

Url = str
SourceName = str
DatetimeIsh = Union[datetime, date]
Context = str
Second = int

# TODO hmm. arguably, source and context are almost same things...
class Loc(NamedTuple):
    title: str
    href: Optional[str]=None

    @classmethod
    def make(cls, title: str, href: Optional[str]=None) -> 'Loc':
        return cls(title=title, href=href)

    @classmethod
    def file(cls, path: PathIsh, line: Optional[int]=None):
        ll = '' if line is None else f':{line}'
        loc = f'{path}{ll}'
        return cls.make(
            title=loc,
            href=f'emacs:{loc}'
        )

    # TODO need some uniform way of string conversion
    # but generally, it would be
    # (url|file)(linenumber|json_path|anchor)


# TODO serialize unions? Might be a bit mad...
class PreVisit(NamedTuple):
    url: Url
    # TODO back to DatetimeIsh, but somehow make compatible to dbcache?
    dt: datetime
    locator: Loc
    context: Optional[Context] = None
    duration: Optional[Second] = None
    # TODO shit. I need to insert it in chrome db....
    # TODO gonna be hard to fill retroactively.
    # spent: Optional[Second] = None
    debug: Optional[str] = None

Visit = PreVisit

Extraction = Union[Visit, Exception]

class DbVisit(NamedTuple):
    norm_url: Url
    orig_url: Url
    dt: datetime
    locator: Loc
    src: Optional[SourceName] = None
    context: Optional[Context] = None
    duration: Optional[Second] = None

    @staticmethod
    def make(p: PreVisit, src: SourceName) -> Res['DbVisit']:
        try:
            if isinstance(p.dt, datetime):
                dt = p.dt
            elif isinstance(p.dt, date):
                # TODO that won't be with timezone..
                dt = datetime.combine(p.dt, datetime.min.time()) # meh..
            else:
                raise AssertionError(f'unexpected date: {p.dt}, {type(p.dt)}')
        except Exception as e:
            return e

        try:
            nurl = canonify(p.url)
        except Exception as e:
            return e

        return DbVisit(
            # TODO shit, can't handle errors properly here...
            norm_url=nurl,
            orig_url=p.url,
            dt=dt,
            locator=p.locator,
            context=p.context,
            duration=p.duration,
            src=src,
        )


Filter = Callable[[Url], bool]

def make_filter(thing) -> Filter:
    if isinstance(thing, str):
        rc = re.compile(thing)
        def filter_(u: str) -> bool:
            return rc.search(u) is not None
        return filter_
    else: # must be predicate
        return thing


def get_logger() -> logging.Logger:
    return logging.getLogger("promnesia")

# TODO do i really need to inherit this??
class History(Sized):
    # TODO I guess instead filter on DbVisit making site?
    FILTERS: List[Filter] = [
        make_filter(f) for f in
        [
            r'^chrome-devtools://',
            r'^chrome-extension://',
            r'^chrome-error://',
            r'^chrome-native://',
            r'^chrome-search://',

            r'chrome://newtab',
            r'chrome://apps',
            r'chrome://history',

            r'^about:',
            r'^blob:',
            r'^view-source:',

            r'^content:',

            # TODO maybe file:// too?
            # chrome-search:
        ]
    ]

    @classmethod
    def add_filter(cls, filterish):
        cls.FILTERS.append(make_filter(filterish))

    def __init__(self, *, src: SourceName):
        self.vmap: Dict[PreVisit, DbVisit] = {}
        # TODO err... why does it map from previsit???
        self.logger = get_logger()
        self.src = src

    # TODO mm. maybe history should get filters from some global config?
    # wonder how okay is it to set class attribute..

    @classmethod
    def filtered(cls, url: Url) -> bool:
        for f in cls.FILTERS:
            if f(url):
                return True
        return False

    @property
    def visits(self) -> List[DbVisit]:
        return list(self.vmap.values())

    def register(self, v: PreVisit) -> Optional[Exception]:
        # TODO should we filter before normalising? not sure...
        if History.filtered(v.url):
            return None

        # TODO perhaps take normalised into account here??
        if v in self.vmap:
            return None

        res = DbVisit.make(v, src=self.src)
        if isinstance(res, CanonifyException):
            self.logger.error('error while canonnifying %s... ignoring', v)
            self.logger.exception(res)
            return None
        elif isinstance(res, Exception):
            return res
        else:
            db_visit = res

        self.vmap[v] = db_visit
        return None
        # TODO hmm some filters make sense before stripping off protocol...

    ## only used in tests?..
    def _nmap(self):
        from itertools import groupby
        key = lambda x: x.norm_url
        return {k: list(g) for k, g in groupby(sorted(self.visits, key=key), key=key)}

    def __len__(self) -> int:
        return len(self._nmap())

    def __contains__(self, url) -> bool:
        return url in self._nmap()

    def __getitem__(self, url: Url):
        return self._nmap()[url]
    #

    def __repr__(self):
        return 'History{' + repr(self.visits) + '}'


# kinda singleton
@lru_cache(1)
def get_tmpdir():
    import tempfile
    tdir = tempfile.TemporaryDirectory(suffix="promnesia")
    return tdir

# TODO use mypy literal?
Syntax = str


@lru_cache(None)
def _get_extractor(syntax: Syntax):
    from urlextract import URLExtract # type: ignore
    u = URLExtract()
    # https://github.com/lipoja/URLExtract/issues/13
    if syntax in {'org', 'orgmode', 'org-mode'}: # TODO remove hardcoding..
        u._stop_chars_right |= {'[', ']'}
        u._stop_chars_left  |= {'[', ']'}
    elif syntax in {'md', 'markdown'}:
        pass
    # u._stop_chars_right |= {','}
    # u._stop_chars_left  |= {','}
    return u


def sanitize(url: str) -> str:
    # TODO not sure it's a really good idea.. but seems easiest now
    # TODO have whitelisted urls that allow trailing parens??
    url = url.strip(',.…\\')
    if 'wikipedia' not in url:
        # urls might end with parens en.wikipedia.org/wiki/Widget_(beer)
        url = url.strip(')')
    return url

# TODO sort just in case? not sure..
def _extract_line_urls(*, s: str, syntax: Syntax) -> List[str]:
    # TODO unit test for escaped urls.. or should it be in normalise instead?
    if len(s.strip()) == 0:
        return [] # optimize just in case..

    # TODO special handling for org links

    # TODO fuck. doesn't look like it's handling multiple urls in same line well...
    # ("python.org/one.html python.org/two.html",
    # hopefully ok... I guess if there are spaces in url we are fucked anyway..
    extractor = _get_extractor(syntax=syntax)
    urls: List[str] = []
    for x in s.split():
        urls.extend(extractor.find_urls(x))

    return [sanitize(u) for u in urls]


def _extract_urls(*, s: str, syntax: Syntax) -> Iterable[Url]:
    for line in s.splitlines():
        yield from _extract_line_urls(s=line, syntax=syntax)


def extract_urls(s: str, syntax: Syntax='') -> List[Url]:
    return list(_extract_urls(s=s, syntax=syntax))


# TODO maybe belongs to HPI?
def from_epoch(ts: int) -> datetime:
    return datetime.fromtimestamp(ts, tz=pytz.utc)


class PathWithMtime(NamedTuple):
    path: Path
    mtime: float

    @classmethod
    def make(cls, p: Path):
        return cls(
            path=p,
            mtime=p.stat().st_mtime,
        )


class Source:
    # TODO make sure it works with empty src?
    # TODO later, make it properly optional?
    def __init__(self, ff, *args, src: SourceName='', name: SourceName='', **kwargs) -> None:
        self.ff = ff
        self.args = args
        self.kwargs = kwargs
        if src is not None:
            warnings.warn("'src' argument is deprecated, please use 'name' instead", DeprecationWarning)
        self.src = name or src

# TODO deprecated
Indexer = Source


# TODO do we really need it?
def previsits_to_history(extractor, *, src: SourceName) -> Tuple[List[DbVisit], List[Exception]]:
    ex = extractor
    # TODO isinstance wrapper?
    # TODO make more defensive?
    logger = get_logger()

    log_info: str
    if isinstance(ex, Indexer):
        log_info = f'{ex.ff.__module__}:{ex.ff.__name__} {ex.args} {ex.kwargs} ...'
        extr = lambda: ex.ff(*ex.args, **ex.kwargs)
    else:
        # TODO if it's a lambda?
        log_info = f'{ex.__module__}:{ex.__name__}'
        extr = ex


    logger.info('extracting via %s ...', log_info)

    h = History(src=src)
    errors = []
    previsits = list(extr()) # TODO DEFENSIVE HERE!!!
    for p in previsits:
        if isinstance(p, Exception):
            errors.append(p)
            parts = ['indexer emitted exception\n']
            # eh, exception type is ignored by format_exception completely, apparently??
            parts.extend(traceback.format_exception(Exception, p, p.__traceback__))
            logger.error(''.join(parts))
            continue

        # TODO check whether it's filtered before construction? probably doesn't really impact
        res = h.register(p)
        if isinstance(res, Exception):
            logger.exception(res)
            errors.append(res)

    # TODO should handle filtering properly?
    logger.info('extracting via %s: got %d visits', log_info, len(h))
    return h.visits, errors


# not sure if necessary anymore?
# NOTE: used in configs...
def last(path: PathIsh, *parts: str) -> Path:
    pp = os.path.join(str(path), *parts)
    return Path(max(glob(pp, recursive=True)))


from .kython.klogging2 import setup_logger

def echain(ex: Exception, cause: Exception) -> Exception:
    ex.__cause__ = cause
    return ex


def slugify(x: str) -> str:
    # https://stackoverflow.com/a/38766141/706389
    import re
    valid_file_name = re.sub(r'[^\w_.)( -]', '', x)
    return valid_file_name




# extracts stuff from hypothes.is json backup
from typing import Iterable

from ..common import Visit, Extraction, get_logger, Loc

import my.hypothesis as hyp


def index() -> Iterable[Extraction]:
    logger = get_logger()

    for h in hyp.get_highlights():
        if isinstance(h, Exception):
            yield h
            continue
        hl = h.highlight
        ann = h.annotation
        cparts = []
        if hl is not None:
            cparts.append(hl)
        if ann is not None:
            cparts.extend(['comment: ' + ann])
        yield Visit(
            url=h.url,
            dt=h.created,
            context='\n\n'.join(cparts),
            locator=Loc.make(
                title='hypothesis',
                href=h.hyp_link,
            )
        )



from contextlib import contextmanager
from pathlib import Path
import re
from time import sleep
from subprocess import Popen, check_output
from typing import Optional, List, Union

# TODO decorator that records a video if a certain env var/flag is set (pass a custom name too)

@contextmanager
def hotkeys(geometry: Optional[str]=None):
    # TODO kill in advance??
    ctx = Popen([
        'screenkey',
        '--no-detach',
        '--timeout', '2',
        '--key-mode', 'composed',
        '--font-size', 'large',
        # TODO font color?
        *([] if geometry is None else ['-g', geometry]),
    ])
    with ctx as p:
        try:
            yield p
        finally:
            p.kill()



@contextmanager
def record(output: Optional[Path]=None, wid: Optional[str]=None, quality: Optional[str]=None):
    assert output is not None, "TODO use tmp file or current dir??"
    # TODO to fullscreen if None?
    assert wid is not None


    # ugh. no idea wtf is happening here... why is position 2,90??
    # wmctrl -i -r 230686723 -e '0,0,0,400,400'
    # xdotool getwindowgeometry 230686723
    # Window 230686723
    #   Position: 2,90 (screen: 0)
    #   Geometry: 400x400
    # Position + Geometry don't add up to the screen size. fuck.
    #
    # ok, xwininfo seems more reliable
    #
    # xwininfo -id $(xdotool getactivewindow)'
    out = check_output(['xwininfo', '-id', wid]).decode('utf8').replace('\n', ' ')
    m = re.search(r'geometry (\d+)x(\d+)[+-](\d+)[+-](\d+)', out)
    assert m is not None, out
    w, h, x, y = m.groups()

    # fuck.
    titlebar = 32

    # fuck x 2
    margin   = 28

    cmd: List[Union[Path, str]] = [
        'ffmpeg',
        '-f', 'x11grab',
        '-y',
        '-r', '30',
        '-s', f'{w}x{titlebar + int(h)}',
        '-i', f':0.0+{x},{margin + int(y)}',
        output,
    ]
    # TODO not sure if need converter script
    # TODO double check there are no ffmpeg processes remaining?
    # maybe, set timeout?

    with Popen(cmd) as p:
        # early check
        sleep(0.5)
        assert p.poll() is None, f'{cmd} died!'

        try:
            yield p
        finally:
            assert p.poll() is None, f'{cmd} died!'

            p.terminate()
            p.wait(timeout=10)


# https://stackoverflow.com/a/52669454/706389
CURSOR_SCRIPT = '''
function enableCursor() {
  var seleniumFollowerImg = document.createElement("img");
  seleniumFollowerImg.setAttribute('src', 'data:image/png;base64,'
    + 'iVBORw0KGgoAAAANSUhEUgAAABQAAAAeCAQAAACGG/bgAAAAAmJLR0QA/4ePzL8AAAAJcEhZcwAA'
    + 'HsYAAB7GAZEt8iwAAAAHdElNRQfgAwgMIwdxU/i7AAABZklEQVQ4y43TsU4UURSH8W+XmYwkS2I0'
    + '9CRKpKGhsvIJjG9giQmliHFZlkUIGnEF7KTiCagpsYHWhoTQaiUUxLixYZb5KAAZZhbunu7O/PKf'
    + 'e+fcA+/pqwb4DuximEqXhT4iI8dMpBWEsWsuGYdpZFttiLSSgTvhZ1W/SvfO1CvYdV1kPghV68a3'
    + '0zzUWZH5pBqEui7dnqlFmLoq0gxC1XfGZdoLal2kea8ahLoqKXNAJQBT2yJzwUTVt0bS6ANqy1ga'
    + 'VCEq/oVTtjji4hQVhhnlYBH4WIJV9vlkXLm+10R8oJb79Jl1j9UdazJRGpkrmNkSF9SOz2T71s7M'
    + 'SIfD2lmmfjGSRz3hK8l4w1P+bah/HJLN0sys2JSMZQB+jKo6KSc8vLlLn5ikzF4268Wg2+pPOWW6'
    + 'ONcpr3PrXy9VfS473M/D7H+TLmrqsXtOGctvxvMv2oVNP+Av0uHbzbxyJaywyUjx8TlnPY2YxqkD'
    + 'dAAAAABJRU5ErkJggg==');
  seleniumFollowerImg.setAttribute('id', 'selenium_mouse_follower');
  seleniumFollowerImg.setAttribute('style', 'position: absolute; z-index: 99999999999; pointer-events: none; left:0; top:0');
  document.body.appendChild(seleniumFollowerImg);
  document.onmousemove = function (e) {
    document.getElementById("selenium_mouse_follower").style.left = e.pageX + 'px';
    document.getElementById("selenium_mouse_follower").style.top  = e.pageY + 'px';
  };
};

enableCursor();
'''


# https://stackoverflow.com/a/987376/706389
SELECT_SCRIPT = '''
function selectText(node) {
    if (document.body.createTextRange) {
        const range = document.body.createTextRange();
        range.moveToElementText(node);
        range.select();
    } else if (window.getSelection) {
        const selection = window.getSelection();
        const range = document.createRange();
        range.selectNodeContents(node);
        selection.removeAllRanges();
        selection.addRange(range);
    } else {
        console.warn("Could not select text in node: Unsupported browser.");
    }
}
'''



from datetime import datetime
from pathlib import Path
from typing import Iterator, NamedTuple, Optional

from ..common import get_logger, Extraction, Url, PathIsh, Res, Visit, echain, Loc


import mistletoe # type: ignore
from mistletoe import Document # type: ignore
from mistletoe.span_token import AutoLink, Link # type: ignore
import mistletoe.block_token as BT # type: ignore
from mistletoe.html_renderer import HTMLRenderer # type: ignore


renderer = HTMLRenderer()


block_tokens = tuple(getattr(BT, name) for name in BT.__all__)


class Parsed(NamedTuple):
    url: Url
    context: Optional[str]


Result = Res[Parsed]


# the fuck...
#
# from mistletoe import Document
# d = Document('''
# # heading
# ## sub
# ## sub2
# ''')
# d.children[0].content
# Out[13]: 'sub2'

# meh, but for now fine I guess
HTML_MARKER = '!html '


def _ashtml(block) -> str:
    res = renderer.render(block)
    if res.startswith('<p>') and res.endswith('</p>'):
        res = res[3: -4] # meh, but for now fine
    return res


class Parser:
    def __init__(self, path: Path):
        self.doc = mistletoe.Document(path.read_text())

    def _extract(self, cur, last_block) -> Iterator[Parsed]:
        if not isinstance(cur, (AutoLink, Link)):
            # hopefully that's all??
            return

        url = cur.target
        # TODO fuck. it doesn't preserve line numbers/positions in text???

        # ugh. It can't output markdown.. https://github.com/miyuchina/mistletoe/issues/4
        context = None if last_block is None else HTML_MARKER + _ashtml(last_block)
        yield Parsed(url=url, context=context)


    def _walk(self, cur, last_block) -> Iterator[Result]:
        if isinstance(cur, block_tokens):
            last_block = cur

        try:
            yield from self._extract(cur, last_block)
        except Exception as e:
            # TODO log context??
            yield e

        children = getattr(cur, 'children', [])
        for c in children:
            yield from self._walk(c, last_block=last_block)


    def walk(self):
        yield from self._walk(self.doc, last_block=None)


def extract_from_file(fname: PathIsh) -> Iterator[Extraction]:
    path = Path(fname)
    fallback_dt = datetime.fromtimestamp(path.stat().st_mtime)

    ex = RuntimeError(f'while extracting from {path}')

    p = Parser(path)
    for r in p.walk():
        if isinstance(r, Exception):
            yield echain(ex, r)
        else:
            yield Visit(
                url=r.url,
                dt=fallback_dt,
                locator=Loc.file(fname), # TODO line number
                context=r.context,
            )



#!/usr/bin/env python3
from datetime import datetime, date
from os.path import join, getsize
from pathlib import Path
from tempfile import TemporaryDirectory
import pytz
from shutil import copytree
import os
from os import mkdir
from os.path import lexists
from typing import Union

import pytest # type: ignore
from pytest import mark # type: ignore


from common import skip_if_ci

from promnesia.common import History, PreVisit
from promnesia.common import Indexer

# TODO need to expire dbcache in tests..

skip = mark.skip


def W(*args, **kwargs):
    if 'src' not in kwargs:
        kwargs['src'] = 'whatever'
    return Indexer(*args, **kwargs)

def history(*args, **kwargs):
    from promnesia.common import previsits_to_history
    return previsits_to_history(*args, **kwargs, src='whatever')[0] # TODO meh


from contextlib import contextmanager
@contextmanager
def extra_path(p: Path):
    import sys
    try:
        sys.path.append(str(p))
        yield
    finally:
        sys.path.pop()

def import_file(p: Union[str, Path], name=None):
    p = Path(p)
    if name is None:
        name = p.stem
    import importlib.util
    spec = importlib.util.spec_from_file_location(name, p) # type: ignore
    foo = importlib.util.module_from_spec(spec)
    with extra_path(p.parent):
        spec.loader.exec_module(foo) # type: ignore
    return foo


# TODO eh. need to separate stuff for history backups out...
backup_db = import_file('scripts/browser_history.py')
populate_db = import_file('scripts/populate-browser-history.py')

def assert_got_tzinfo(visits):
    for v in visits:
        assert v.dt.tzinfo is not None


# TODO I guess global get_config methods is ok? command line can populate it, also easy to hack in code?
# TODO cache should be in the configuration I suppose?

@pytest.fixture
def adhoc_config(tmp_path):
    tdir = Path(tmp_path)
    cdir = tdir / 'cache'
    cdir.mkdir()

    from promnesia import config

    try:
        config.instance = config.Config(
            INDEXERS=[],
            OUTPUT_DIR=tdir,
            CACHE_DIR=cdir,
        )
        yield
    finally:
        config.reset()


@pytest.mark.skip(reason='TODO support unpacked directories in HPI')
def test_takeout_directory(adhoc_config, tmp_path):
    from my.cfg import config
    class user_config:
        takeout_path = 'testdata/takeout'
    config.google = user_config
    import promnesia.sources.takeout as tex

    visits = history(W(tex.index))
    assert len(visits) > 0 # kinda arbitrary?

    assert_got_tzinfo(visits)



def test_with_error():
    class ExtractionError(Exception):
        pass
    def err_ex():
        for i in range(3):
            if i == 1:
                yield ExtractionError()
            else:
                yield PreVisit(
                    url=f'http://test{i}',
                    dt=datetime.utcfromtimestamp(0),
                    locator=None,
                )
    hist = history(lambda: err_ex())
    assert len(hist) == 2


def test_takeout_new_zip(adhoc_config):
    from my.cfg import config
    class user_config:
        takeout_path = 'testdata/takeout-20150518T000000Z.zip'
    config.google = user_config

    import promnesia.sources.takeout as tex
    visits = history(tex.index)
    assert len(visits) == 3
    [vis] = [v for v in visits if v.norm_url == 'takeout.google.com/settings/takeout']

    edt = datetime(
        year=2018,
        month=9,
        day=18,
        hour=5,
        minute=48,
        second=23,
        tzinfo=pytz.utc,
    )
    assert vis.dt == edt

    assert_got_tzinfo(visits)


# TODO run condition?? and flag to force all
@skip_if_ci("TODO try triggering firefox on CI? not sure if that's possible...")
def test_chrome(tmp_path):
    from promnesia.sources.browser import chrome
    tdir = Path(tmp_path)

    path = tdir / 'history'
    populate_db.merge_from('chrome', from_=None, to=path)
    # TODO hmm, it actually should be from merged db....

    hist = history(W(chrome, path))
    assert len(hist) > 10 # kinda random sanity check

    assert_got_tzinfo(hist)


@skip_if_ci("TODO try triggering firefox on CI? not sure if that's possible...")
def test_firefox(tmp_path):
    tdir = Path(tmp_path)
    path = backup_db.backup_history('firefox', to=tdir, profile='*release*')
    # shouldn't fail at least

        # [hist] = list(chrome_gen.iter_chrome_histories(path, 'sqlite'))
        # assert len(hist) > 10 # kinda random sanity check

        # render([hist], join(tdir, 'res.json'))

        # assert_got_tzinfo(hist)


def test_plaintext_path_extractor():
    import promnesia.sources.shellcmd as custom_gen
    from promnesia.sources.plaintext import extract_from_path

    visits = history(W(custom_gen.index,
        extract_from_path('testdata/custom'),
    ))
    assert {
        v.orig_url for v in visits
    } == {
        'http://google.com',
        'http://google.com/',
        'http://some-weird-domain/whatever',
        'https://google.com',
        'http://what.about.this.link',
    }

# TODO perhaps it belongs to canonify?
def test_normalise():
    import promnesia.sources.shellcmd as custom_gen
    from promnesia.sources.plaintext import extract_from_path

    visits = history(W(custom_gen.index,
        extract_from_path('testdata/normalise'),
    ))
    assert len(visits) == 7
    assert {
        v.norm_url for v in visits
    } == {
        'hi.com',
        'reddit.com/post',
        'argos.co.uk/webapp/wcs/stores/servlet/OrderItemDisplay',
        'youtube.com/watch?v=XXlZfc1TrD0',
        'youtube.com/watch?v=XXlZfc1Tr11',
    }


def test_normalise_weird():
    import promnesia.sources.shellcmd as custom_gen
    from promnesia.sources.plaintext import extract_from_path

    visits = history(W(
        custom_gen.index,
        extract_from_path('testdata/weird.txt'),
    ))
    norms = {v.norm_url for v in visits}

    # TODO assert there are no spaces in the database?
    assert "urbandictionary.com/define.php?term=Belgian%20Whistle" in norms
    assert "en.wikipedia.org/wiki/Dinic%27s_algorithm" in norms


@skip("use a different way to specify filter other than class variable..")
def test_filter():
    import promnesia.sources.shellcmd as custom_gen
    from promnesia.sources.plaintext import extract_from_path

    History.add_filter(r'some-weird-domain')
    hist = custom_gen.get_custom_history(
        extract_from_path('testdata/custom'),
    )
    assert len(hist) == 4 # chrome-error got filtered out

def test_custom():
    import promnesia.sources.shellcmd as custom_gen

    hist = history(W(custom_gen.index,
        """grep -Eo -r --no-filename '(http|https)://\S+' testdata/custom""",
    ))
    # TODO I guess filtering of equivalent urls should rather be tested on something having context (e.g. org mode)
    assert len(hist) == 5



TESTDATA_CHROME_HISTORY = "/L/data/promnesia/testdata/chrome-history"

def get_chrome_history_backup(td: str):
    copytree(TESTDATA_CHROME_HISTORY, join(td, 'backup'))

@skip("TODO move this to populate script instead")
def test_merge():
    merge = backup_db.merge

    # TODO third is implicit... use merging function
    with TemporaryDirectory() as tdir:
        get_chrome_history_backup(tdir)
        first  = join(tdir, "backup/20180415/History")
        second = join(tdir, "backup/20180417/History")

        mdir = join(tdir, 'merged')
        mkdir(mdir)
        merged_path = join(mdir, 'merged.sql')


        def merged_size() -> int:
            return getsize(merged_path)

        merge(merged_path, first)
        fsize = merged_size()

        merge(merged_path, first)
        fsize_2 = merged_size()

        assert fsize == fsize_2

        merge(merged_path, second)
        ssize = merged_size()

        assert ssize > fsize

        merge(merged_path, second)
        ssize_2 = merged_size()

        assert ssize_2 == ssize


def _test_merge_all_from(tdir):
    merge_all_from = backup_db.merge_all_from # type: ignore
    mdir = join(tdir, 'merged')
    mkdir(mdir)
    mfile = join(mdir, 'merged.sql')

    get_chrome_history_backup(tdir)

    merge_all_from(mfile, join(tdir, 'backup'), None)

    first  = join(tdir, "backup/20180415/History")
    second = join(tdir, "backup/20180417/History")

    # should be removed
    assert not lexists(first)
    assert not lexists(second)

    import promnesia.sources.chrome as chrome_ex

    hist = history(W(chrome_ex.extract, mfile))
    assert len(hist) > 0

    older = hist['github.com/orgzly/orgzly-android/issues']
    assert any(v.dt.date() < date(year=2018, month=1, day=17) for v in older)
    # in particular, "2018-01-16 19:56:56"

    newer = hist['en.wikipedia.org/wiki/Notice_and_take_down']
    assert any(v.dt.date() >= date(year=2018, month=4, day=16) for v in newer)

    # from implicit db
    newest = hist['feedly.com/i/discover']
    assert any(v.dt.date() >= date(year=2018, month=9, day=27) for v in newest)

@skip("TODO move this to populate script")
def test_merge_all_from(tmp_path):
    tdir = Path(tmp_path)
    _test_merge_all_from(tdir)
    # TODO and also some other unique thing..

if __name__ == '__main__':
    pytest.main([__file__])

# TODO mark tests with @skip_if_ci

# TODO once I integrate test for db population, test chrome/firefox extractors



from typing import Iterator, Optional

from ..common import Extraction, get_logger, Visit, Loc

from my.pocket import get_articles


def index() -> Iterator[Extraction]:
    # TODO use docstring from my. module? E.g. describing which pocket format is expected

    for a in get_articles():
        # TODO not sure if can make more specific link on pocket?
        loc = Loc.make(title='pocket', href=a.pocket_link)
        hls = a.highlights
        if len(hls) == 0:
            yield Visit(
                url=a.url,
                dt=a.added,
                context=None,
                locator=loc,
            )
        for hl in hls:
            yield Visit(
                url=a.url,
                dt=hl.created,
                context=hl.text,
                locator=loc,
            )



from pathlib import Path
from typing import Iterator, Optional

from ..common import Extraction, get_logger, Visit, Loc, PathIsh, extract_urls

from my.fbmessenger import messages


# TODO rename Extraction to smth shorter? even Result is good enough?
def index() -> Iterator[Extraction]:
    # TODO add db path?
    for m in messages():
        text = m.text
        if text is None:
            continue
        urls = extract_urls(text)
        if len(urls) == 0:
            continue

        # TODO m.author would be niceneeds to be implemented in fbmessenger model
        loc = Loc.make(
            title=f'chat with {m.thread.name}',
            # eh, not all threads have nicknames, and not sure how to extract reliably
            href=f'https://www.messenger.com/t/{m.thread.thread_id}',
        )
        for u in urls:
            yield Visit(
                url=u,
                dt=m.dt,
                context=m.text,
                locator=loc,
            )




from ..common import get_logger, get_tmpdir, PathIsh

from pathlib import Path
import os.path
import os

from shlex import quote

# https://linux-and-mac-hacks.blogspot.co.uk/2013/04/use-grep-and-regular-expressions-to.html
_URL_REGEX = r'\b(https?|ftp|file)://[-A-Za-z0-9+&@#/%?=~_|!:,.;]*[-A-Za-z0-9+&@#/%=~_|]'

# -n to output line numbers so we could restore context
# -I to ignore binaries
_GREP_CMD = r"""grep --color=never -Eo -I {grep_args} --exclude-dir=".git" '{regex}' {path} || true"""


def _extract_from_dir(path: str) -> str:
    return _GREP_CMD.format(
        grep_args="-r -n",
        regex=_URL_REGEX,
        path=path, # TODO quote here too?
    )

def _extract_from_file(path: str) -> str:
    return _GREP_CMD.format(
        grep_args="-n",
        regex=_URL_REGEX,
        path=f"{quote(path)} /dev/null", # dev null to trick into displaying filename
    )

# TODO eh, misleading..
def extract_from_path(path: PathIsh) -> str:
    pp = Path(path)

    tdir = get_tmpdir()

    logger = get_logger()
    if pp.is_dir(): # TODO handle archives here???
        return _extract_from_dir(str(pp))
    else:
        if any(pp.suffix == ex for ex in (
                '.xz',
                '.bz2',
                '.gz',
                '.zip',
        )):
            logger.info(f"Extracting from compressed file {path}")
            raise RuntimeError()
            import lzma
            from tempfile import NamedTemporaryFile
            # TODO hopefully, no collisions
            fname = os.path.join(tdir.name, os.path.basename(path))
            with open(fname, 'wb') as fo:
                with lzma.open(path, 'r') as cf:
                    fo.write(cf.read())
                return _extract_from_file(fname)
        else:
            return _extract_from_file(str(pp))



from promnesia.common import extract_urls

def test_extract():
    lines = """
 I've enjoyed [Chandler Carruth's _There Are No Zero-cost Abstractions_](
 https://www.youtube.com/watch?v=rHIkrotSwcc) very much.
""".strip()
    assert set(extract_urls(lines)) == {'https://www.youtube.com/watch?v=rHIkrotSwcc'}


def test_extract_2():
   text = '''‍♂️ Чтобы снизить вероятность ошибиться, важно знать про когнитивные искажения.
   Если для вас это новое словосочетание, начните с книжки
   "Гарри Поттер и Методы рационального мышления" - http://hpmor.ru/, если вы знакомы с понятием - читайте цепочки на сайтах
   lesswrong.ru и lesswrong.com, книжку Даниэля Канемана "Thinking, fast and slow" и канал Пион https://t.me/ontologics
   '''
   assert set(extract_urls(text)) == {'http://hpmor.ru/', 'lesswrong.ru', 'lesswrong.com', 'https://t.me/ontologics'}

def test_extract_md():
    lines = '''
Hey, I recently implemented a new extension for that [addons.mozilla.org](https://addons.mozilla.org/en-US/firefox/addon/org-grasp-for-org-capture/), [github](https://github.com/karlicoss/grasp), perhaps it could be useful for you!
    '''
    assert set(extract_urls(lines)) == {
        'addons.mozilla.org',
        'https://addons.mozilla.org/en-US/firefox/addon/org-grasp-for-org-capture/',
        'https://github.com/karlicoss/grasp',
    }



from pathlib import Path
from typing import Optional, Union, Iterable, TypeVar
from urllib.parse import unquote # TODO mm, make it easier to rememember to use...

import dataset # type: ignore

from ..common import PathIsh, PreVisit, get_logger, Loc, extract_urls, from_epoch, Extraction, echain

# TODO potentially, belongs to my. package

# TODO kython?
T = TypeVar('T')
def unwrap(res: Union[T, Exception]) -> T:
    if isinstance(res, Exception):
        raise res
    else:
        return res


# TODO move to common?
def dataset_readonly(db: Path):
    # see https://github.com/pudo/dataset/issues/136#issuecomment-128693122
    import sqlite3
    creator = lambda: sqlite3.connect(f'file:{db}?mode=ro', uri=True)
    return dataset.connect('sqlite:///' , engine_kwargs={'creator': creator})


def index(database: PathIsh) -> Iterable[Extraction]:
    logger = get_logger()

    path = Path(database)
    assert path.is_file()

    # TODO context manager?
    db = dataset_readonly(path)

    def make_query(text_query: str):
        return f"""
WITH entities AS (
   SELECT 'dialog' as type, id, coalesce(username, id) as handle, coalesce(first_name || " " || last_name, username, id) as display_name FROM users
   UNION
   SELECT 'group' as type, id, id as handle                    , coalesce(name, id) as display_name FROM chats
)
SELECT src.display_name AS chatname
     , src.handle       AS chat
     , snd.display_name AS sender
     , M.time           AS time
     , {text_query}     AS text
     , M.id             AS mid
FROM messages AS M
                                                                    /* chat types are 'dialog' (1-1), 'group' and 'supergroup' */
                                                                    /* this is abit hacky way to handle all groups in one go */
LEFT JOIN entities AS src    ON M.source_id = src.id AND src.type = (CASE M.source_type WHEN 'supergroup' THEN 'group' ELSE M.source_type END)
LEFT JOIN entities AS snd    ON M.sender_id = snd.id AND snd.type = 'dialog'
WHERE
    M.message_type NOT IN ('service_message', 'empty_message')
/* used to do this, but doesn't really give much of a speedup */
/* AND (M.has_media == 1 OR (text LIKE '%http%')) */
ORDER BY time;
    """.strip()

    # TODO yield error if chatname or chat or smth else is null?
    for row in db.query(make_query('M.text')):
        try:
            yield from _handle_row(row)
        except Exception as ex:
            yield echain(RuntimeError(f'While handling {row}'), ex)
            # , None, sys.exc_info()[2]
            # TODO hmm. traceback isn't preserved; wonder if that's because it's too heavy to attach to every single exception object..

    # old (also 'stable') version doesn't have 'json' column yet...
    if 'json' in db['messages'].columns:
        for row in db.query(make_query("json_extract(json, '$.media.webpage.description')")):
            try:
                yield from _handle_row(row)
            except Exception as ex:
                yield echain(RuntimeError(f'While handling {row}'), ex)


def _handle_row(row) -> Iterable[Extraction]:
    text = row['text']
    if text is None:
        return
    urls = extract_urls(text)
    if len(urls) == 0:
        return
    dt            = from_epoch(row['time'])
    mid: str      = unwrap(row['mid'])

    # TODO perhaps we could be defensive with null sender/chat etc and still emit the Visit
    sender: str   = unwrap(row['sender'])
    chatname: str = unwrap(row['chatname'])
    chat: str     = unwrap(row['chat'])

    in_context = f'https://t.me/{chat}/{mid}'
    for u in urls:
        # https://www.reddit.com/r/Telegram/comments/6ufwi3/link_to_a_specific_message_in_a_channel_possible/
        # hmm, only seems to work on mobile app, but better than nothing...
        yield PreVisit(
            url=unquote(u),
            dt=dt,
            context=f"{sender}: {text}",
            locator=Loc.make(
                title=f"chat with {chatname}",
                href=in_context,
            ),
        )



#!/usr/bin/env python3
# TODO perhaps make it external script?
import argparse
from pathlib import Path
import logging
import sys
from typing import Dict, List, Any, NamedTuple, Optional, Iterator, Set, Tuple


from .common import DbVisit, Url, PathWithMtime # TODO ugh. figure out pythonpath
from .kython import kompress

# TODO include latest too?
# from cconfig import ignore, filtered

def get_logger():
    return logging.getLogger('promnesia-db-changes')

# TODO return error depending on severity?


from typing import TypeVar, Sequence


T = TypeVar('T')

def eliminate_by(sa: Sequence[T], sb: Sequence[T], key):
    def make_dict(s: Sequence[T]) -> Dict[str, List[T]]:
        res: Dict[str, List[T]] = {}
        for a in s:
            k = key(a)
            ll = res.get(k, None)
            if ll is None:
                ll = []
                res[k] = ll
            ll.append(a)
        return res
    da = make_dict(sa)
    db = make_dict(sb)
    ka = set(da.keys())
    kb = set(db.keys())
    onlya: Set[T] = set()
    common: Set[T] = set()
    onlyb: Set[T] = set()
    for k in ka.union(kb):
        la = da.get(k, [])
        lb = db.get(k, [])
        common.update(la[:min(len(la), len(lb))])
        if len(la) > len(lb):
            onlya.update(la[len(lb):])
        if len(lb) > len(la):
            onlyb.update(lb[len(la):])

    return onlya, common, onlyb


def compare(before: List[DbVisit], after: List[DbVisit], between: str, *, log=True) -> List[DbVisit]:
    logger = get_logger()
    logger.info('comparing between: %s', between)

    errors: List[DbVisit] = []

    umap: Dict[Url, List[DbVisit]] = {}
    for a in after:
        url = a.norm_url
        xx = umap.get(url, []) # TODO canonify here?
        xx.append(a)
        umap[url] = xx

    def reg_error(b):
        errors.append(b)
        if log:
            logger.error('between %s missing %s', between, b)
            print('ignoreline "%s", # %s %s' % ('exid', b.norm_url, b.src), file=sys.stderr)


    # the idea is that we eliminate items simultaneously from both sets
    eliminations = [
        ('identity'               , lambda x: x),
        ('without dt'             , lambda x: x._replace(src='', dt='')),
        ('without context'        , lambda x: x._replace(src='',        context='', locator='')),
        ('without dt and context' , lambda x: x._replace(src='', dt='', context='', locator='')),
    ]
    for ename, ekey in eliminations:
        logger.info('eliminating by %s', ename)
        logger.info('before: %d, after: %d', len(before), len(after))
        before, common, after = eliminate_by(before, after, key=ekey)
        logger.info('common: %d, before: %d, after: %d', len(common), len(before), len(after))

    logger.info('removing explicitly ignored items')
    # before = filtered(before, between=between, umap=umap)
    logger.info('before: %d', len(before))

    for b in before:
        reg_error(b)

    return errors

def setup_parser(p):
    # TODO better name?
    p.add_argument('--intermediate-dir', type=Path)
    p.add_argument('--last', type=int, default=2)
    p.add_argument('--all', action='store_const', const=0, dest='last')
    p.add_argument('paths', nargs='*')


def get_files(args):
    if len(args.paths) == 0:
        int_dir = args.intermediate_dir
        assert int_dir.exists()
        files = list(sorted(int_dir.glob('*.sqlite*')))
        files = files[-args.last:]
    else:
        files = [Path(p) for p in args.paths]
    return files


def main():
    p = argparse.ArgumentParser()
    setup_parser(p)
    args = p.parse_args()
    files = get_files(args)

    errors = list(compare_files(*files))
    if len(errors) > 0:
        sys.exit(1)


def compare_files(*files: Path, log=True) -> Iterator[Tuple[str, DbVisit]]:
    assert len(files) > 0

    logger = get_logger()
    logger.info('comparing %s', files)

    last = None
    last_dts = None
    for f in files:
        logger.info('processing %r', f)
        name = f.name
        this_dts = name[0: name.index('.')] # can't use stem due to multiple extensions..

        from promnesia.server import _get_stuff # TODO ugh
        engine, binder, table = _get_stuff(PathWithMtime.make(f))

        with engine.connect() as conn:
            vis = [binder.from_row(row) for row in conn.execute(table.select())]

        if last is not None:
            between = f'{last_dts}:{this_dts}'
            errs = compare(last, vis, between=between, log=log)
            for e in errs:
                yield between, e
        last = vis
        last_dts = this_dts

if __name__ == '__main__':
    main()




import os
from functools import wraps
from pathlib import Path

import pytest # type: ignore


def under_ci():
    return 'CI' in os.environ


def skip_if_ci(reason):
    return pytest.mark.skipif(under_ci(), reason=reason)


def uses_x(f):
    @skip_if_ci('Uses X server')
    @wraps(f)
    def ff(*args, **kwargs):
        return f(*args, **kwargs)
    return ff


@pytest.fixture
def tdir(tmp_path):
    yield Path(tmp_path)


def tdata(path: str) -> Path:
    pp = Path(__file__).parent.parent / 'testdata'
    assert pp.is_dir()
    return pp.absolute() / path





#!/usr/bin/env python3
import argparse
from datetime import datetime
import logging
import tempfile
from os.path import lexists
from typing import Optional, NamedTuple, Sequence, Tuple, Union
from pathlib import Path
import subprocess
from subprocess import check_call, DEVNULL, check_output

# pip3 install python-magic
import magic # type: ignore
mime = magic.Magic(mime=True)

from browser_history import Browser, backup_history, CHROME, FIREFOX, guess_db_date


def get_logger():
    return logging.getLogger('populate-browser-history')


def sqlite(db: Path, script, method=check_call, **kwargs):
    return method(['sqlite3', str(db), script], **kwargs)


def entries(db: Path) -> Optional[int]:
    if not db.exists():
        return None
    return int(sqlite(db, 'SELECT COUNT(*) FROM visits', method=check_output).decode('utf8'))


Col = Union[str, Tuple[str, Optional[str]]] # tuple is renaming
ColType = str


class Schema(NamedTuple):
    cols: Sequence[Tuple[Col, ColType]]
    key: Sequence[str]


SchemaCheck = Tuple[str, str]

def create(db: Path, table: str, schema: Schema):
    things = []
    for cc, tp in schema.cols:
        from_: str
        to: Optional[str]
        if isinstance(cc, str):
            from_ = cc
            to = cc
        else:
            (from_, to) = cc
        if to is not None:
            to = to.split('.')[-1] # strip off table alias
            to = to.split(' AS ')[-1]
            things.append(f'{to} {tp}')

    query = f"""
CREATE TABLE {table}(
    {', '.join(things)},
    PRIMARY KEY ({', '.join(schema.key)})
);
    """
    sqlite(db, query)


# at first, I was merging urls and visits tables separately... but it's kinda messy when you e.g. reinstall OS and table ids reset
# so joining before inserting makes a bit more sense.. we're throwing id anyway since it's fairly useless for the same reasons
# TODO it's a bit slow now because of the repeating joins presumably... could pass last handled visit date or something... not sure if it's safe
# TODO not sure how to make chunk read only?
def merge_browser(
        merged: Path,
        chunk: Path,
        schema: Schema,
        schema_check: SchemaCheck,
        query: str,
):
    check_table, expected = schema_check
    # ugh. a bit ugly but kinda works
    res = sqlite(chunk, f"select group_concat(name, ', ') from pragma_table_info('{check_table}')", method=check_output).decode('utf8').strip()
    if res != expected and not res in expected:
        raise AssertionError(f'expected schema {expected}, got {res}')
    # TODO default??


    if not merged.exists():
        create(merged, 'visits', schema)

    proj = ', '.join(c for c, _ in schema.cols) # type: ignore
    query = f"""
ATTACH '{chunk}' AS chunk;

INSERT OR IGNORE INTO main.visits
    SELECT {proj}
    {query};

DETACH chunk;
    """
    sqlite(merged, query)


class Extr(NamedTuple):
    detector: str
    schema_check: SchemaCheck
    schema: Schema
    query: str


chrome = Extr(
    detector='keyword_search_terms',
    schema_check=(
        'visits', [
            'visits', "id, url, visit_time, from_visit, transition, segment_id, visit_duration, incremented_omnibox_typed_score",
            'visits', "id, url, visit_time, from_visit, transition, segment_id, visit_duration",
        ]
    ),
    schema=Schema(cols=[
        ('U.url'                                  , 'TEXT'),

        # while these two are not very useful, might be good to have just in case for some debugging
        ('U.id AS urlid', 'INTEGER'),
        ('V.id AS vid', 'INTEGER'),

        ('V.visit_time'                             , 'INTEGER NOT NULL'),
        ('V.from_visit'                             , 'INTEGER'),
        ('V.transition'                             , 'INTEGER NOT NULL'),
        # V.segment_id looks useless
        ('V.visit_duration'                         , 'INTEGER NOT NULL'),
        # V.omnibox thing looks useless
    ], key=('url', 'visit_time', 'vid', 'urlid')),
    query='FROM chunk.visits as V, chunk.urls as U WHERE V.url = U.id',
)

# https://www.forensicswiki.org/wiki/Mozilla_Firefox_3_History_File_Format#moz_historyvisits
firefox = Extr(
    detector='moz_meta',
    schema_check=('moz_historyvisits', "id, from_visit, place_id, visit_date, visit_type, session"),
    schema=Schema(cols=[
        ('P.url'       , 'TEXT'),

        ('P.id AS pid' , 'INTEGER'),
        ('V.id AS vid' , 'INTEGER'),

        ('V.from_visit', 'INTEGER'),
        ('V.visit_date', 'INTEGER'),
        ('V.visit_type', 'INTEGER'),
        # not sure what session is form but could be useful?..
        ('V.session'   , 'INTEGER'),
    ], key=('url', 'visit_date', 'vid', 'pid')),
    query='FROM chunk.moz_historyvisits as V, chunk.moz_places as P WHERE V.place_id = P.id',
)


firefox_phone = Extr(
    detector='remote_devices',
    schema_check=('visits', "_id, history_guid, visit_type, date, is_local"),
    schema=Schema(cols=[
        ('H.url'         , 'TEXT NOT NULL'),

        ('H.guid AS guid', 'TEXT'),
        ('H._id  AS hid'  , 'INTEGER'),
        ('V._id  AS vid'  , 'INTEGER'),

        ('V.visit_type'  , 'INTEGER NOT NULL'),
        ('V.date'        , 'INTEGER NOT NULL'),
        # ('is_local'    , 'INTEGER NOT NULL'),
    ], key=('url', 'date', 'vid', 'hid')),
    query='FROM chunk.visits as V, chunk.history as H  WHERE V.history_guid = H.guid',
)

def merge(merged: Path, chunk: Path):
    logger = get_logger()
    logger.info(f"Merging {chunk} into {merged}")
    if lexists(merged):
        logger.info("DB size before: %s items %d bytes", entries(merged), merged.stat().st_size)
    else:
        logger.info(f"DB doesn't exist yet: {merged}")

    candidates = []
    for ff in [chrome, firefox, firefox_phone]:
        res = sqlite(chunk, f"SELECT * FROM {ff.detector}", method=subprocess.run, stdout=DEVNULL, stderr=DEVNULL)
        if res.returncode == 0:
            candidates.append(ff)
    assert len(candidates) == 1
    merger = candidates[0]

    merge_browser(merged=merged, chunk=chunk, schema=merger.schema, schema_check=merger.schema_check, query=merger.query)
    logger.info("DB size after : %s items %d bytes", entries(merged), merged.stat().st_size)


def merge_from(browser: Optional[Browser], from_: Optional[Path], to: Path, profile='*'):
    assert not to.is_dir()

    logger = get_logger()
    with tempfile.TemporaryDirectory() as td:
        tdir = Path(td)

        if from_ is None:
            assert browser is not None
            backup_history(browser, tdir, profile=profile)
            from_ = tdir
        else:
            assert from_.exists()

        is_db = lambda x: x.is_file() and mime.from_file(str(x)) in ['application/x-sqlite3']

        if is_db(from_):
            files = [from_]
        else:
            files = sorted(x for x in from_.rglob('*') if is_db(x))

        for dbfile in files:
            # TODO maybe, assert they all of the same type?
            logger.info('merging %s', dbfile)
            merge(merged=to, chunk=dbfile)


def _helper(tmp_path, browser, profile='*'):
    logger = get_logger()

    tdir = Path(tmp_path)
    merged = tdir / 'merged.sqlite'

    entr = entries(merged)
    assert entr is None

    merge_from(browser, None, merged, profile=profile)
    merge_from(browser, None, merged, profile=profile)

    entr = entries(merged)
    assert entr is not None
    assert entr > 100 # quite arbitrary, but ok for now

def test_merge_chrome(tmp_path):
    _helper(tmp_path, CHROME)

def test_merge_firefox(tmp_path):
    _helper(tmp_path, FIREFOX, profile='*release')


def main():
    logger = get_logger()

    p = argparse.ArgumentParser()
    p.add_argument('--browser', type=Browser, required=False)
    p.add_argument('--to', type=Path, required=True)
    p.add_argument('--from', type=Path, default=None)
    p.add_argument('--profile', default='*')
    args = p.parse_args()

    from_ = getattr(args, 'from')

    # TODO need to mark already handled? although it's farily quick as it s
    # maybe should use the DB thing to handle merged??
    merge_from(browser=args.browser, from_=from_, to=args.to, profile=args.profile)


if __name__ == '__main__':
    main()





from contextlib import contextmanager
from datetime import datetime
import json
import os
from pathlib import Path
from shutil import copy
import signal
from subprocess import check_output, check_call, Popen, PIPE
import time
from typing import NamedTuple, ContextManager

import pytz

from integration_test import index_hypothesis, index_urls
from common import tdir, under_ci


class Helper(NamedTuple):
    port: str


# TODO use proper random port
TEST_PORT = 16556


def next_port():
    global TEST_PORT
    TEST_PORT += 1
    return TEST_PORT


@contextmanager
def tmp_popen(*args, **kwargs):
    with Popen(*args, **kwargs, preexec_fn=os.setsid) as p:
        try:
            yield p
        finally:
            # ugh. otherwise was getting orphaned children...
            os.killpg(os.getpgid(p.pid), signal.SIGTERM)

# meh
def promnesia_bin(*args):
    # not sure it's a good idea to diverge, but not sure if there's a better way either?
    if under_ci():
        # should be able to use the installed version
        return ['promnesia', *args]
    else:
        # use version from the repository
        root = Path(__file__).parent.parent
        pm = root / 'scripts/promnesia'
        return [pm, *args]


@contextmanager
def wserver(db: Path): # TODO err not sure what type should it be... -> ContextManager[Helper]:
    port = str(next_port())
    cmd = [
        'serve',
        '--quiet',
        '--port', port,
        '--db'  , str(db),
    ]
    with tmp_popen(promnesia_bin(*cmd)) as server:
        print("Giving few secs to start server up")
        time.sleep(3)
        print("Started server up, db: {db}".format(db=db))

        yield Helper(port=port)

        print("DONE!!!!")


@contextmanager
def _test_helper(tmp_path):
    tdir = Path(tmp_path)
    cache_dir = tdir / 'cache'
    cache_dir.mkdir()

    # TODO extract that into index_takeout?
    # TODO ugh. quite hacky...
    template_config = Path(__file__).parent.parent / 'testdata' / 'test_config.py'
    copy(template_config, tdir)
    config = tdir / 'test_config.py'
    with config.open('a') as fo:
        fo.write(f"""
OUTPUT_DIR = '{tdir}'
CACHE_DIR  = '{cache_dir}'
""")

    check_call(promnesia_bin('index', '--config', config))

    with wserver(db=tdir / 'promnesia.sqlite') as srv:
        yield srv


def post(*args):
    cmd = [
        'http',
        # '--timeout', '10000', # useful for debugging
        '--ignore-stdin',
        'post',
        *args,
    ]
    return json.loads(check_output(cmd).decode('utf8'))


def test_query_instapaper(tdir):
    index_hypothesis(tdir)
    test_url = "http://www.e-flux.com/journal/53/59883/the-black-stack/"
    with wserver(db=tdir / 'promnesia.sqlite') as helper:
        response = post(f'http://localhost:{helper.port}/visits', f'url={test_url}')
        assert len(response['visits']) > 5
        # TODO actually test response?


def test_visits(tmp_path):
    test_url = 'https://takeout.google.com/settings/takeout'
    with _test_helper(tmp_path) as helper:
        for q in range(3):
            print(f"querying {q}")
            response = post(f'http://localhost:{helper.port}/visits', f'url={test_url}')
            assert len(response['visits']) == 1


def test_search(tdir):
    index_hypothesis(tdir)
    test_url = "http://www.e-flux.com"
    with wserver(db=tdir / 'promnesia.sqlite') as helper:
        response = post(f'http://localhost:{helper.port}/search', f'url={test_url}')
        assert len(response['visits']) == 8


def test_visited(tmp_path):
    test_url = 'https://takeout.google.com/settings/takeout'
    with _test_helper(tmp_path) as helper:
        response = post(f'http://localhost:{helper.port}/visited', f"""urls:=["{test_url}","http://badurl.org"]""")
        assert response == [True, False]


def test_search_around(tmp_path):
    tdir = Path(tmp_path)
    index_hypothesis(tdir)
    dt = pytz.utc.localize(datetime.strptime("2017-05-22T10:58:14.082375", '%Y-%m-%dT%H:%M:%S.%f'))
    test_ts = int(dt.timestamp())
    # test_ts = int(datetime(2016, 12, 13, 12, 31, 4, 229275, tzinfo=pytz.utc).timestamp())
    # TODO hmm. perhaps it makes more sense to run query in different process and server in main process for testing??
    with wserver(db=tdir / 'promnesia.sqlite') as helper:
        response = post(f'http://localhost:{helper.port}/search_around', f'timestamp={test_ts}')
        # TODO highlight original url in extension??
        assert 5 < len(response['visits']) < 20


# TODO right.. I guess that triggered because of reddit indexer specifically
# TODO could probably reuse parts of tests to be both integration/server and end2end when necessary?
def test_visits_hier(tdir):
    test_url = 'https://www.reddit.com/r/QuantifiedSelf/comments/d6m7bd/android_app_to_track_and_export_application_use/'
    urls = {
        test_url: 'parent url',
        'https://reddit.com/r/QuantifiedSelf/comments/d6m7bd/android_app_to_track_and_export_application_use/f0vem56': 'Some context',
        'https://reddit.com/r/QuantifiedSelf/comments/d6m7bd/android_app_to_track_and_export_application_use/whatever': None, # no context so should be ignored..
    }
    indexer = index_urls(urls)
    indexer(tdir)
    with wserver(db=tdir / 'promnesia.sqlite') as helper:
        response = post(f'http://localhost:{helper.port}/visits', f'url={test_url}')
        assert {v['context'] for v in response['visits']} == {'parent url', 'Some context'}


def test_status(tdir):
    with wserver(db='/does/not/exist') as helper:
        response = post(f'http://localhost:{helper.port}/status')
        assert response['db'] == None # defensive, it doesn't exist
        version = response['version']
        assert version is not None
        assert len(version.split('.')) >= 2 # random check..



from datetime import datetime
from html.parser import HTMLParser
from pathlib import Path
from typing import Iterable, List, Tuple

from ..common import PathIsh, Visit, Loc, Extraction

# TODO present error summary in the very end; import errors -- makes sense to show 

# TODO on some exceptions, request a fallback to text?

# TODO not sure if import should be inside extract_from_file or here...
from bs4 import BeautifulSoup # type: ignore[import]

def extract_from_file(fname: PathIsh) -> Iterable[Extraction]:

    ts = datetime.fromtimestamp(Path(fname).stat().st_mtime)
    # TODO just allow passing file as timestamp?

    soup = BeautifulSoup(Path(fname).read_text(errors='replace'), 'lxml')
    for a in soup.find_all('a'):
        href = a.attrs.get('href')
        if href is None:
            # TODO ignore #?
            continue
        text = a.text

        yield Visit(
            url=href,
            dt=ts,
            locator=Loc.file(fname),
            context=text,
        )


# eh. it's not really going to work well because of nesting inside <a> tags..
# class Parser(HTMLParser):
#     results: List[Tuple[str, str]] = []
#
#     link = None
#     title = None
#
#     def handle_starttag(self, tag, attrs):
#         if tag != 'a':
#             return
#         href = dict(attrs).get('href')
#         if href is None:
#             return
#         self.link = href
#
#     def handle_data(self, data):
#         # TODO only if url is set??
#         # also, only set once?
#         if self.link is not None:
#             breakpoint()
#         # TODO what if it's bold?
#         title = 'sup'
#
#     def handle_endtag(self, tag):
#         if tag != 'a':
#             return
#         # TODO if old isn't none, emit
#         if self.link is not None:
#             self.links.append((self.link, self.title))
#         self.link = None




from itertools import chain
from typing import Iterable

from ..common import Visit, Loc, extract_urls, Extraction, get_logger


from my.reddit import Submission, Comment, Save, Upvote
from my.reddit import submissions, comments, saved, upvoted


def index() -> Iterable[Extraction]:
    logger = get_logger()
    # TODO should probably use markdown parser here?

    logger.info('processing saves')
    for s in saved():
        try:
            yield from _from_save(s)
        except Exception as e:
            yield e

    logger.info('processing comments')
    for c in comments():
        try:
            yield from _from_comment(c)
        except Exception as e:
            yield e

    logger.info('processing submissions')
    for sub in submissions():
        try:
            yield from _from_submission(sub)
        except Exception as e:
            yield e

    logger.info('processing upvotes')
    for u in upvoted():
        try:
            yield from _from_upvote(u)
        except Exception as e:
            yield e


def _from_save(i: Save) -> Iterable[Extraction]:
    locator = Loc.make(
        title='Reddit save',
        href=i.url,
    )
    yield from _from_common(i, locator=locator)


def _from_comment(i: Comment) -> Iterable[Extraction]:
    locator = Loc.make(
        title='Reddit comment',
        href=i.url,
    )
    yield from _from_common(i, locator=locator)


def _from_submission(i: Submission) -> Iterable[Extraction]:
    locator = Loc.make(
        title=f'Reddit submission: {i.title}',
        href=i.url,
    )
    yield from _from_common(i, locator=locator)


def _from_upvote(i: Upvote) -> Iterable[Extraction]:
    locator = Loc.make(
        title=f'Reddit upvote',
        href=i.url,
    )
    yield from _from_common(i, locator=locator)


def _from_common(i: 'RedditThing', locator: Loc) -> Iterable[Extraction]:
    urls = [i.url]
    # TODO this should belong to HPI.. fix permalink handling I guess
    # ok, it's not present for all of them..
    lurl = i.raw.get('link_url')
    if lurl is not None:
        urls.append(lurl)
    lurl = i.raw.get('url')
    if lurl is not None:
        urls.append(lurl)

    for url in chain(urls, extract_urls(i.text)):
        yield Visit(
            url=url,
            dt=i.created,
            context=i.text,
            locator=locator,
        )


from datetime import datetime
import typing
if typing.TYPE_CHECKING:
    # TODO define in rexport?
    from typing_extensions import Protocol
    from typing import Dict, Any
    class RedditThing(Protocol):
        @property
        def raw(self) -> Dict[str, Any]: ...
        @property
        def url(self) -> str: ...
        @property
        def created(self) -> datetime: ...
        @property
        def text(self) -> str: ...

# TODO should probably use markdown link parser throughout the module



from typing import Iterator, Optional

from ..common import Extraction, get_logger, Visit, Loc, PathIsh, extract_urls

import my.twitter.all as tw


def get(o, k):
    if isinstance(o, dict):
        return o[k]
    else:
        return getattr(o, k)


def index() -> Iterator[Extraction]:
    logger = get_logger()
    # TODO hmm. tweets themselves are sort of visits? not sure if they should contribute..
    processed = 0
    for t in tw.tweets():
        processed += 1
        try:
            urls = t.urls
        except Exception as e: # just in case..
            yield e
            urls = []

        if len(urls) == 0:
            # if entities haven't detected anything it usually means RT or reply in my case, so worth trying again to extract
            # e.g. replies from json twitter takeouts don't seem to have entities set
            urls = extract_urls(t.text)
            # t.co refers to the retweeted tweet, so perhaps not very meaningful
            urls = [u for u in urls if '/t.co/' not in u]

        loc = Loc.make(title='twitter', href=t.permalink)
        for u in urls:
            yield Visit(
                url=u,
                dt=t.created_at,
                context=t.text,
                locator=loc,
            )
    logger.info('processed %d tweets', processed)


# ok, so it doesn't necessarily have everything in entities, eg.
# {
#   "retweeted" : false,
#   "source" : "<a href=\"http://twitter.com\" rel=\"nofollow\">Twitter Web Client</a>",
#   "entities" : {
#     "hashtags" : [ ],
#     "symbols" : [ ],
#     "user_mentions" : [ ],
#     "urls" : [ ]
#   },
#   "full_text" : "http://old.slackware.ru/article.ghtml?ID=544  Забавно =)",
#  ...
# }



from pathlib import Path

from promnesia.common import Source
from promnesia.sources.plaintext import extract_from_path
import promnesia.sources.shellcmd as shellcmd # type: ignore
import promnesia.sources.takeout as takeout # type: ignore


class Sources:
    @staticmethod
    def TAKEOUT():
        class user_config:
            takeout_path = 'testdata/takeout-20150518T000000Z.zip'

        import my.config
        my.config.google = user_config

        return Source(
            takeout.index,
            name='takeout',
        )

    PLAIN = Source(
        shellcmd.index,
        extract_from_path('testdata/custom'),
        name='test',
    )


SOURCES = [
    Sources.PLAIN,
    Sources.TAKEOUT,
]



#!/usr/bin/env python3
from datetime import datetime
from pathlib import Path
from subprocess import check_output
from typing import Optional
import filecmp
import logging

Browser = str

CHROME = 'chrome'
FIREFOX = 'firefox'

def get_logger():
    return logging.getLogger('browser-history')


# TODO kython?
# TODO the with key?
def only(it):
    values = list(it)
    if len(values) == 1:
        return values[0]
    raise RuntimeError(f'Expected a single value: {values}')


def get_path(browser: Browser, profile: str='*') -> Path:
    if browser == 'chrome':
        bpath = Path('~/.config/google-chrome').expanduser()
        dbs = bpath.glob(profile + '/History')
    elif browser == 'firefox':
        bpath = Path('~/.mozilla/firefox/').expanduser()
        dbs = bpath.glob(profile + '/places.sqlite')
    else:
        raise RuntimeError(f'Unexpected browser {browser}')
    ldbs = list(dbs)
    if len(ldbs) == 1:
        return ldbs[0]
    raise RuntimeError(f'Expected single database, got {ldbs}. Perhaps you want to use --profile argument?')



def test_get_path():
    get_path('chrome')
    get_path('firefox', profile='*-release')


def atomic_copy(src: Path, dest: Path):
    """
    Supposed to handle cases where the file is changed while we were copying it.
    """
    import shutil

    differs = True
    while differs:
        res = shutil.copy(src, dest)
        differs = not filecmp.cmp(str(src), str(res))


def format_dt(dt: datetime) -> str:
    return dt.strftime('%Y%m%d%H%M%S')


def backup_history(browser: Browser, to: Path, profile: str='*', pattern=None) -> Path:
    assert to.is_dir()
    logger = get_logger()

    now = format_dt(datetime.utcnow())

    path = get_path(browser, profile=profile)

    pattern = path.stem + '-{}' + path.suffix if pattern is None else pattern
    fname = pattern.format(now)


    res = to / fname
    logger.info('backing up to %s', res)
    # if your chrome is open, database would normally be locked, so you can't just make a snapshot
    # so we'll just copy it till it converge. bit paranoid, but should work
    atomic_copy(path, res)
    logger.info('done!')
    return res


def test_backup_history(tmp_path):
    tdir = Path(tmp_path)
    backup_history(CHROME, tdir)
    backup_history(FIREFOX, tdir, profile='*-release')


def guess_db_date(db: Path) -> str:
    maxvisit = check_output([
        'sqlite3',
        '-csv',
        db,
        'SELECT max(datetime(((visits.visit_time/1000000)-11644473600), "unixepoch")) FROM visits;'
    ]).decode('utf8').strip().strip('"');
    return format_dt(datetime.strptime(maxvisit, "%Y-%m-%d %H:%M:%S"))


def test_guess(tmp_path):
    tdir = Path(tmp_path)
    db = backup_history(CHROME, tdir)
    guess_db_date(db)


def main():
    logger = get_logger()
    import argparse
    p = argparse.ArgumentParser()
    p.add_argument('--browser', type=Browser, required=True)
    p.add_argument('--profile', type=str, default='*', help='Use to pick the correct profile to back up. If unspecified, will assume a single profile')
    p.add_argument('--to', type=Path, required=True)
    args = p.parse_args()

    # TODO do I need pattern??
    backup_history(browser=args.browser, to=args.to, profile=args.profile)


if __name__ == '__main__':
    main()



"""
Some experimental ideas on JSON processing.
This is a bit overengineered and I admit it!
I'll make it more readable, but in the meantime feel free to open an issue if you're confused about something.
"""

from typing import Any, Dict, List, Union, Tuple, cast


JDict = Dict[str, Any] # TODO not sure if we can do recursive..
JList = List[Any]
JPrim = Union[str, int, float] # , type(None)]

Json = Union[JDict, JList, JPrim]

JPathPart = Tuple[Json, Union[str, int]]

JPath = Tuple[JPathPart, ...]


class JsonProcessor:
    SKIP = object()

    def handle_dict(self, js: JDict, jp: JPath) -> Any:
        pass

    def handle_list(self, js: JList, jp: JPath) -> Any:
        pass

    def handle_str(self, js: str, jp: JPath) -> Any:
        pass

    def do_dict(self, js: JDict, jp: JPath) -> None:
        # pylint: disable=assignment-from-no-return
        res = self.handle_dict(js, jp)
        if res is self.SKIP:
            return
        for k, v in js.items():
            path = cast(JPath, jp + ((js, k), ))
            self._do(v, path)

    def do_list(self, js: JList, jp: JPath) -> None:
        # pylint: disable=assignment-from-no-return
        res = self.handle_list(js, jp)
        if res is self.SKIP:
            return
        for i, x in enumerate(js):
            path = cast(JPath, jp + ((js, i), ))
            self._do(x, path)

    def _do(self, js: Json, path: JPath) -> None:
        if isinstance(js, dict): # TODO have functions for dict like, list like etc
            self.do_dict(js, path)
        elif isinstance(js, list):
            self.do_list(js, path)
        elif isinstance(js, str):
            self.handle_str(js, path)
        elif isinstance(js, (int, bool, float, type(None))):
            pass # TODO process that as well
        else:
            raise RuntimeError(f'unexpected item {js} of type {type(js)}')

    def run(self, js: Json) -> None:
        path = cast(JPath, ())
        self._do(js, path)

    @classmethod
    def kpath(cls, path: JPath) -> Tuple[JPathPart, ...]:
        return tuple(x[1] for x in path) # type: ignore

# TODO path is a sequence of jsons and keys?

def test_json_processor():
    handled = []
    class Proc(JsonProcessor):
        def handle_dict(self, value: JDict, path):
            if 'skipme' in self.kpath(path):
                return JsonProcessor.SKIP

        def handle_str(self, value: str, path):
            if 'http' in value:
                handled.append((value, path))

    j = {
        'skipme': {
            'x': {
                'y': [
                    123,
                    {
                        'description': 'whatever',
                        'link': 'http://ya.ru',
                    },
                ]
            }
        },
        'a': [1, 2, 3],
        'x': {
            'y': [
                123,
                {
                    'description': 'whatever',
                    'link': 'http://reddit.com',
                },
            ]
        },
    }

    p = Proc()
    p.run(j)
    assert len(handled) > 0

    [h1] = handled
    (link, path) = h1
    assert link == 'http://reddit.com'
    pp = [p[1] for p in path]
    assert pp == ['x', 'y', 1, 'link']


if __name__ == '__main__':
    test_json_processor()




from pathlib import Path
import re
import json
from time import sleep

# TODO share with grasp... maybe move to kython?
def open_extension_page(driver, page: str):
    ff = {
        'chrome' : get_extension_page_chrome,
        'firefox': get_extension_page_firefox,
    }[driver.name]
    extension_prefix = ff(driver=driver)
    driver.get(extension_prefix + '/' + page)


# TODO looks like it used to be posssible in webdriver api?
# at least as of 2011 https://github.com/gavinp/chromium/blob/681563ea0f892a051f4ef3d5e53438e0bb7d2261/chrome/test/webdriver/test/chromedriver.py#L35-L40
# but here https://github.com/SeleniumHQ/selenium/blob/master/cpp/webdriver-server/command_types.h there are no Extension commands
# also see driver.command_executor._commands
def get_extension_page_chrome(driver):
    chrome_profile = Path(driver.capabilities['chrome']['userDataDir'])
    prefs_file = chrome_profile / 'Default/Preferences'
    # there are some default extensions as well (e.g. cloud print)
    # also oddly enough user install extensions don't have manifest information, so we can't find it by name

    # seems to be quite a bit asynchronous (e.g. up to several seconds), so good to be defensive for a bit
    prefs = None
    addon_id = None
    for _ in range(30):
        sleep(0.5)
        if not prefs_file.exists():
            continue

        prefs = json.loads(prefs_file.read_text())
        settings = prefs.get('extensions', {}).get('settings', None)
        if settings is None:
            continue

        addon_ids = [k for k, v in settings.items() if Path(v['path']).name == f'extension_{k}']
        if len(addon_ids) == 0:
            print('addon_ids is none')
            continue
        [addon_id] = addon_ids
    assert addon_id is not None
    assert prefs is not None
    return f'chrome-extension://{addon_id}'


def get_extension_page_firefox(driver):
    moz_profile = Path(driver.capabilities['moz:profile'])
    prefs_file = moz_profile / 'prefs.js'

    # doesn't appear immediately after installing somehow, so need to wait for a bit..
    for _ in range(10):
        sleep(0.5)
        if prefs_file.exists():
            break

    addon_name = 'temporary_addon'
    # TODO ok, apparently I should add it to tips on using or something..
    addon_name = 'promnesia@karlicoss.github.com'

    addon_id = None
    for line in prefs_file.read_text().splitlines():
        # temporary-addon\":\"53104c22-acd0-4d44-904c-22d11d31559a\"}")
        m = re.search(addon_name + r'.....([0-9a-z-]+)."', line)
        if m is None:
            continue
        addon_id = m.group(1)
    assert addon_id is not None
    return f'moz-extension://{addon_id}'

# TODO could also check for errors



from contextlib import contextmanager
from datetime import timedelta, datetime
from pathlib import Path
from time import sleep
from subprocess import check_call
from typing import Optional


from common import uses_x
from end2end_test import FF, CH, browsers, _test_helper
from end2end_test import PYTHON_DOC_URL
from integration_test import index_urls
from end2end_test import confirm
from end2end_test import configure, get_window_id

from record import record, hotkeys, CURSOR_SCRIPT, SELECT_SCRIPT


def real_db():
    from private import real_db_path, test_filter
    from tempfile import TemporaryDirectory
    import shutil
    def indexer(tdir: Path):
        tdb = tdir / 'promnesia.sqlite'
        # tdb.touch()
        shutil.copy(real_db_path, tdb)

        filt = test_filter()
        if filt is not None:
            check_call([
                'sqlite3',
                tdb,
                filt,
            ])

    return indexer


class Annotator:
    def __init__(self):
        self.start = datetime.now()
        self.l = []

    def __call__(self, *args, **kwargs) -> None:
        return self.annotate(*args, **kwargs)

    def annotate(self, text: str, length=2) -> None:
        # TODO how to display during recording??
        now = datetime.now()
        print(f"ANNOTATING: {text}")
        self.l.append((now, text, length))

    def build(self, **extra):
        from pysubs2 import SSAFile, SSAEvent, Color # type: ignore[import]
        millis = lambda td: td / timedelta(milliseconds=1)
        subs = (
            SSAEvent(
                start=millis(t - self.start),
                end  =millis(t - self.start + timedelta(seconds=length)),
                text=text.replace('\n', r'\N'), # \N necessary for SSA files
            ) for t, text, length in self.l
        )
        sf = SSAFile()
        style = sf.styles['Default'].copy()
        style.fontsize = 16 # default is 20, bit too much??
        style.outlinecolor = Color(0, 0, 0, 50) # semitransparent
        style.shadow = 0.0

        style.outline = 0.1
        style.borderstyle = 3 # no idea why 3, but it makes the background apperar in conjunction with outline

        for k, v in extra.items():
            setattr(style, k, v)
        sf.styles['Default'] = style
        for s in subs:
            sf.append(s)
        return sf.to_string('ass')



@contextmanager
def demo_helper(*, tmp_path, browser, path: Path, indexer=real_db, before=None, subs_position='topleft', size='40%', **kwargs):
    # TODO literal type??
    # https://stackoverflow.com/a/25880038/706389
    # 789
    # 456
    # 123
    spos = {
        'topleft'    : 7,
        'bottomleft' : 1,
        'bottomright': 3,
    }[subs_position]

    position = f'''
.promnesia {{
    --right: 1;
    --size:  {size};
    background-color: rgba(236, 236, 236, 0.8);
}}
    '''
    # TODO hmm. not sure if the whole sidebar should have a background instead??

    with _test_helper(tmp_path, indexer(), None, browser=browser) as helper:
        driver = helper.driver
        driver_wid = get_window_id(driver)

        # full display resolution
        W = 2560
        H = 1440

        margin   = 28 # make up space for panel etc... figure that out properly later
        titlebar = 32
        # TODO I think firefox counts position as the actual top left corner, but geometry without the title bar???

        w = W // 2
        h = H - margin - titlebar
        def set_geometry(wid: str):
            check_call([
                'wmctrl',
                '-i',
                '-r', wid,
                '-e', f'0,0,{margin},{w},{h}',
            ])
        set_geometry(wid=driver_wid)

        extras = kwargs
        if 'highlights' not in extras:
            extras['highlights'] = False

        configure(
            driver,
            host=None, port=None, # TODO meh
            notification=False,
            position=position,
            **extras,
        )

        driver.get('about:blank')

        if before is not None:
            before(driver)

        # TODO eh. not sure if this geometry is consistent with ffmpeg...
        geometry = f'{w}x400+0+{h - 400}'
        with hotkeys(geometry=geometry):
            # TODO record directly in webm? but need to set quality
            rpath = path.with_suffix('.mp4')
            with record(rpath, wid=driver_wid):
                ann = Annotator()
                yield helper, ann

            subs = path.with_suffix('.ssa')
            subs.write_text(ann.build(alignment=spos))
            out  = path.with_suffix('.webm')

            converter = Path(__file__).parent.absolute() / 'convert_screencast.py'
            check_call([
                converter,
                path,
            ])

            # TODO unlink subs here?
            # subs.unlink()


# TODO use ass?
# https://trac.ffmpeg.org/wiki/HowToBurnSubtitlesIntoVideo


demo = False


def prompt(what: str):
    if demo:
        return
    confirm(what)


fast = False

def wait(x):
    if fast:
        return
    print(f"Sleeping for {x} seconds")
    sleep(x)


@uses_x
@browsers(FF, CH)
def test_demo_quick(tmp_path, browser):
    path = Path('demos/quick')
    with demo_helper(
            tmp_path=tmp_path,
            browser=browser,
            path=path,
            subs_position='topleft',
    ) as (helper, annotate):
        driver = helper.driver

        driver.get('https://en.wikipedia.org/wiki/Whatever_(slang)')
        annotate('''
One two three four five.
Six seven eight nine ten?
        ''', length=5)
        wait(5)


demos = Path('demos')


# TODO need to determine that uses X automatically
@uses_x
@browsers(FF, CH)
def test_demo_mark_visited(tmp_path, browser):
    # TODO wonder if it's possible to mess with settings in local storage? unlikely...

    path = demos / 'mark-visited'

    # TODO fast mode??
    url = 'https://slatestarcodex.com/'
    with demo_helper(
            tmp_path=tmp_path,
            browser=browser,
            path=path,
            subs_position='bottomright',
    ) as (helper, annotate):
        driver = helper.driver

        driver.get(url)

        # TODO display subs nicer??

        annotate('''
On the left you can see a blogroll with recommended blogs.
Lots of sites there!
''', length=3)

        wait(3)

        annotate('''
You feel like reading something new.
Which are the ones you haven't seen before?
        ''', length=3)

        wait(3)

        # TODO request focus on 'prompt'??
        # prompt('continue?')

        helper.mark_visited()
        wait(2)

        # TODO somehow mark images with the same annotations??
        helper.screenshot(path.with_suffix('.png'))

        annotate('''
The command marks links you've already visited with dots.
This way you don't have to search your browser history all over for each of them!
        ''', length=5)
        wait(5)

        annotate('''
You can click right on the ones you haven't seen before and start exploring!
        ''', length=8)
        wait(8)


@uses_x
@browsers(FF, CH)
def test_demo_mark_visited_2(tmp_path, browser):
    path = demos / 'mark-visited-2'

    # TODO maybe test on Baez instead?
    # TODO scroll to ?
    url = 'https://www.lesswrong.com/posts/vwqLfDfsHmiavFAGP/the-library-of-scott-alexandria#IV__Medicine__Therapy__and_Human_Enhancement'

    def before(driver):
        driver.get(url)

    with demo_helper(tmp_path=tmp_path, browser=browser, path=path, before=before) as (helper, annotate):
        driver = helper.driver
       
        wait(2)

        annotate('''
Lots of links to explore on this page.
Which ones haven't I seen before?
''', length=5)
        wait(5)

        helper.mark_visited()
        annotate('''
Hotkey press...
        ''', length=1.5)
        wait(1.5)

        annotate('''
Links I've already visited are marked with dots!
        ''', length=8)
        wait(2)
        helper.screenshot(path.with_suffix('.png'))

        wait(6)


# TODO perhaps make them independent of network? Although useful for demos

@uses_x
@browsers(FF, CH)
def test_demo_child_visits(tmp_path, browser):
    path = Path('demos/child-visits')
    with demo_helper(
            tmp_path=tmp_path,
            browser=browser,
            path=path,
            subs_position='bottomleft',
    ) as (helper, annotate):
        driver = helper.driver
        driver.get('https://twitter.com/michael_nielsen/status/1162502843921600512')

        annotate('''
While browsing Twitter, I see an account recommendation.
''', length=4)
        wait(4)

        annotate('''
I really value Michael Nielsen's opinion, so it's worth checking out.
''', length=5)
        wait(5) # TODO maybe, wait by default??

        # TODO move_to, etc. animate!

        driver.get('https://twitter.com/eriktorenberg')

        wait(3)

        # TODO wait till loaded??

        # TODO turn contexts notification on here?
        annotate('''
See the green eye icon in top right?
It means I have run into this account before!
        ''', length=5)
        wait(5)

        wait(1)
        # TODO make hotkey popup larger...
        annotate('''
Let's see...
        ''', length=2)
        wait(1)

        helper.activate()
        wait(2)

        helper.screenshot(path.with_suffix('.png'))

        annotate('''
Right, I've already bookmarked something interesting from that guy before.
I guess I should follow him!
        ''', length=7)
        wait(7)

        # TODO this could also demonstrate jump to the tweet? ( ->->  )



@uses_x
@browsers(FF, CH)
def test_demo_child_visits_2(tmp_path, browser):
    path = Path('demos/child-visits-2')

    def before(driver):
        # jeez. medium takes, like, 15 seconds to load
        driver.get('https://medium.com/@justlv/how-to-build-a-brain-interface-and-why-we-should-connect-our-minds-35003841c4b7')

    with demo_helper(
            tmp_path=tmp_path,
            browser=browser,
            path=path,
            subs_position='bottomleft',
            before=before,
    ) as (helper, annotate):
        driver = helper.driver
        wait(2)

        annotate('''
I ran into this cool post on Hackernews.
Usually I'd also check out the author's blog for more content.
''', length=6)
        wait(6)

        driver.get('https://medium.com/@justlv')
        wait(1)

        annotate('''
The icon is green.
So I've interacted with the page before!
''', length=5)
        wait(5)

        wait(1)
        # TODO make hotkey popup larger...
        annotate('''
Let's see...
        ''', length=2)
        wait(2)

        helper.activate()
        wait(2)

        helper.switch_to_sidebar()

        driver.execute_script(CURSOR_SCRIPT)
        # TODO move to helper??

        tweet = driver.find_element_by_class_name('locator')
        helper.move_to(tweet)

        annotate('''
Cool, I've even tweeted about one of the posts on this blog before!
        ''', length=5)
        wait(5)

        helper.screenshot(path.with_suffix('.png'))

        # TODO original tweet -> smth else??
        annotate('''
Clicking on 'context' will bring me straight to the original tweet.
        ''', length=6)
        wait(2)

        a_tweet = tweet.find_element_by_tag_name('a')
        helper.move_to(a_tweet)

        wait(2)

        a_tweet.click()

        wait(8)


from selenium import webdriver # type: ignore


def scroll_to_text(driver, text: str):
    # ActionChain doesn't work if the element isn't visible. ugh.
    # https://stackoverflow.com/questions/3401343/scroll-element-into-view-with-selenium
    # so have to resort to JS

    element = driver.find_element_by_xpath(f"//*[contains(text(), '{text}')]")

    loc = element.location
    y = loc['y']

    driver.execute_script(f'window.scrollTo(0, {y})')
    # TODO a bit of wait??
   

from end2end_test import get_webdriver


@uses_x
@browsers(FF, CH)
def test_demo_highlights(tmp_path, browser):
    assert browser == FF, browser # because of the profile_dir hack
    path = demos / 'highlights'


    def before(driver):
        from private import instapaper_cookies

        # necessary to set cookies on instapaper..
        driver.get('http://instapaper.com')
        for cookie in instapaper_cookies():
            driver.add_cookie(cookie)

        driver.get('https://instapaper.com/read/1257588750')


    with demo_helper(
            tmp_path=tmp_path,
            browser=browser,
            path=path,
            subs_position='bottomleft',
            highlights=True,
            before=before,
    ) as (helper, annotate):
        # TODO is it possible to disable extension first??
        driver = helper.driver

        wait(2)

        annotate('''
I'm using Instapaper to read and highlight articles while I'm offline on my phone.
        ''', length=7)
        wait(2)

        scroll_to_text(driver, "where things stood")
        wait(3)
        scroll_to_text(driver, "As impossible as it sounds")
        wait(3)

        # TODO go to div class="source" -> a class="original"
        # driver without the extension
        ORIG = 'http://nautil.us/issue/66/clockwork/haunted-by-his-brother-he-revolutionized-physics-rp'
        with get_webdriver(browser, extension=False) as driver2:
            driver2.get(ORIG)

            # TODO maybe, have annotation 'start' and 'interrupt'?
            annotate('''
But if you open the original article, you can't see the annotations!
            ''', length=7)
            wait(2)

            scroll_to_text(driver2, "where things stood")
            wait(3)
            scroll_to_text(driver2, "As impossible as it sounds")
            wait(3)

        annotate('''
Let's try it with Promnesia!
        ''', length=7)
        wait(2)

        driver.get(ORIG)
        scroll_to_text(driver, "where things stood")
        wait(3)
        scroll_to_text(driver, "As impossible as it sounds")
        wait(3)

        annotate('''
Highlights are displayed within the original page!
        ''', length=5)
        wait(5)

        helper.activate()
        wait(2)

        helper.screenshot(path.with_suffix('.png'))

        annotate('''
It works with any highlight source, be it Pocket, Hypothes.is or anything else.
        ''', length=5)
        wait(5)

        annotate('''
Let me demonstrate...
        ''')
        wait(2)

        driver.get('https://en.wikipedia.org/wiki/Empty_Spaces#Composition')
        wait(3)

        helper.activate()
        wait(2)

        # TODO move cursor to the note?
        annotate('''
This clipping is in my org-mode notes!
It's not using any annotation service -- it's just a plaintext file!
        ''', length=7)
        wait(7)


# TODO https://www.youtube.com/watch?v=YKLpz025vYY
# TODO https://www.youtube.com/watch?v=17Q0tJZcsnY
# TODO https://en.wikipedia.org/wiki/SnapPea

@uses_x
@browsers(FF, CH)
def test_demo_how_did_i_get_here(tmp_path, browser):
    path = demos / 'how_did_i_get_here'
    with demo_helper(
            tmp_path=tmp_path,
            browser=browser,
            path=path,
            subs_position='bottomleft',
    ) as (helper, annotate):
        driver = helper.driver
        driver.get('https://www.amazon.co.uk/Topoi-Categorial-Analysis-Logic-Mathematics/dp/0486450260')

        annotate('''
I found this link in my bookmarks.
Hmmm, can't remember, why I added it.
        ''', length=6)
        wait(6)

        helper.activate()
        wait(2)

        annotate('''
If you click on a timestamp, you'll jump straight to the visit in your timeline.
        ''', length=6)
        wait(2)

        helper.switch_to_sidebar()

        driver.execute_script(CURSOR_SCRIPT)
        last_dt = list(driver.find_elements_by_class_name('datetime'))[-1]
        helper.move_to(last_dt)
        wait(4)

        # TODO helper to click 'the' link on the element??
        # last_dt.find_element_by_tag_name('a').click()
        # TODO highlight clicks??
        last_dt.click()
        driver.switch_to.window(driver.window_handles[-1])
        driver.execute_script(CURSOR_SCRIPT)
        wait(1)

        annotate('''
Looks like right before clicking on the Amazon link, I was reading
a book on math.ucr.edu/home/baez/planck
Let's see...
        ''', length=8)
        wait(4)
        # meh..

        node8 = 'math.ucr.edu/home/baez/planck/node8.html'
        baez = driver.find_element_by_link_text(node8)
        helper.move_to(baez)
        wait(4)

        helper.screenshot(path.with_suffix('.png'))

        # driver.execute_script(f"window.open('{node8}')")
        # TODO encapsulate??
        baez.click()
        # driver.switch_to.window(driver.window_handles[-1])

        wait(2)
        annotate('''
Aha! Here's the reference to the book.
        ''', length=5)
        wait(2)
        driver.execute_script(
            # meh
            SELECT_SCRIPT + "\n" + "selectText(document.getElementsByTagName('dd')[17])"
        )
        wait(5)
    # TODO mention that it's chrome history! so I wouldn't have it in my browser history in firefox



@uses_x
@browsers(FF, CH)
def test_demo_watch_later(tmp_path, browser):
    path = demos / 'watch_later'

    def before(driver):
        driver.get('https://www.youtube.com/watch?v=DvT-O_DI4t4')


    with demo_helper(
            tmp_path=tmp_path,
            browser=browser,
            path=path,
            subs_position='bottomleft',
            before=before,
    ) as (helper, annotate):
        driver = helper.driver
        wait(2)

        annotate('''
I have this video in my "Watch later" playlist.
I wonder why?
        ''', length=4)
        wait(4)

        helper.activate()
        wait(2)

        annotate('''
Aha, my friend sent it to me on Telegram.
The sidebar also embeds the message that mentions it.
        ''', length=5)
        wait(5)

        annotate('''
Once I watched it, I want to discuss it with the friend.
        ''', length=3)
        wait(3)

        helper.switch_to_sidebar()
        driver.execute_script(CURSOR_SCRIPT)
        loc = driver.find_element_by_class_name('locator')
        helper.move_to(loc)

        helper.screenshot(path.with_suffix('.png'))

        annotate('''
Clicking on the "chat with" link will jump straight into that message in Telegram web app.
        ''', length=5)
        wait(2)
        a_loc = loc.find_element_by_tag_name('a')
        helper.move_to(a_loc)
        wait(3)

        # TODO not actually clicking, becaus I have no idea how to censor Telegram into a proper demo
        # document.getElementsByClassName('im_dialogs_col')[0].style.display = 'none'
        # TODO get url

        # TODO how to censor the link???

        # driver.get('https://web.telegram.org')
        # breakpoint()
        #
        # TODO hide im_message_from_photo
        # .im_message_body {
        #     color:
        #     white;
        # }
        # TODO <span class="peer_initials nocopy im_message_from_photo user_bgcolor_4" data-content="initials_here"></span>



from pathlib import Path
import re
from subprocess import check_call, run
from typing import Iterable

from ..common import Extraction, PathIsh, get_tmpdir, slugify, get_logger


def index(path: PathIsh, *args, **kwargs) -> Iterable[Extraction]:
    logger = get_logger()
    url = str(path)

    # TODO better context name
    tp = Path(get_tmpdir().name) / slugify(url)

    # TODO careful, set some hard limit on data size? use --quota?
    # https://www.linuxjournal.com/content/downloading-entire-web-site-wget

    cmd = [
        'wget', '--directory-prefix', str(tp),
        '--no-verbose',
        '--recursive',
        '-A', 'html,html,txt', # TODO eh, ideally would use mime type I guess...
        '--no-parent',
        url,
    ]
    # TODO follow sitemap? e.g. gwern
    logger.info(' '.join(cmd))
    res = run(cmd)

    if res.returncode == 8:
        # man wget: 8 means server error (e.g. broken link)
        yield RuntimeError('Encountered server error(s) during downloading')
    else:
        # rest of the errors are a bit more critical..
        res.check_returncode()

    def replacer(p: PathIsh, prefix=str(tp), url=url) -> str:
        ps = str(p)
        pos = ps.find(prefix)
        if pos == -1:
            return ps
        rest = ps[pos + len(prefix):]
        # now this should look kinda like /domain.tld/rest (due to the way wget downloads stuff)
        rest = re.sub(r'/.*?/', '/', rest)
        return url + rest

    # TODO create a file that maps prefix?
    # TODO ugh. it creates a directory with a domain... how to map it to http/https properly?

    # TODO smarter html handling
    from . import auto
    yield from auto.index(tp, *args, replacer=replacer, **kwargs)



from datetime import datetime
from subprocess import check_call, check_output
from typing import Iterator, Optional
from pathlib import Path
from urllib.parse import unquote

import pytz

from ..common import Visit, Loc

def index(command: str) -> Iterator[Visit]:
    output = check_output(command, shell=True)
    lines = [line.decode('utf-8') for line in output.splitlines()]
    # TODO eh?
    for line in lines:
        # TODO wtf is that??? use extract_url??
        protocols = ['file', 'ftp', 'http', 'https']
        for p in protocols:
            split_by = ':' + p + '://'
            if split_by in line:
                parts = line.split(split_by)
                break
        else:
            parts = [line]

        fname: Optional[str]
        lineno: Optional[int]
        url: str
        if len(parts) == 1:
            fname = None
            lineno = None
            url = parts[0]
        else:
            [fname, linenos] = parts[0].rsplit(':', maxsplit=1)
            lineno = int(linenos)
            url = split_by[1:] + parts[1]

        # TODO is it really necessary with locator?
        context = f"{fname}:{lineno}" if fname and lineno else None

        # TODO not sure if even necessary? not that I use canonify
        url = unquote(url)

        ts: datetime
        loc: Loc
        if fname is not None:
            ts = datetime.fromtimestamp(Path(fname).stat().st_mtime)
            loc = Loc.file(fname, line=lineno)
        else:
            # TODO fallback tz??
            ts = datetime.utcnow().replace(tzinfo=pytz.utc)
            loc = Loc.make(command)
        # TODO !1 extract org notes properly...
        yield Visit(
            url=url,
            dt=ts,
            locator=loc,
            context=context,
        )



import csv
import sqlite3
from datetime import datetime
from subprocess import check_output
from typing import Dict, Iterator, List, NamedTuple, Optional, Set
from urllib.parse import unquote

import pytz
from sqlalchemy import Column, MetaData, Table, create_engine # type: ignore

from ..common import Loc, PathIsh, PreVisit, get_logger, Second


def browser_extract(histfile: PathIsh, cols, row_handler) -> Iterator[PreVisit]:
    logger = get_logger()
    logger.debug(f'extracing history from {histfile}')

    # TODO fuck. why doesn't that work???
    # engine = create_engine('sqlite:///{histfile}', echo=True)
    # meta = MetaData()
    # visits = Table('visits', meta, autoload=True, autoload_with=engine)
    # TODO contextmanager
    conn = sqlite3.connect(str(histfile))

    for row in conn.execute(f"SELECT {', '.join(cols)} FROM visits"):
        pv = row_handler(*row)
        yield pv

    logger.debug('done extracing')


def _firefox(cols, histfile: PathIsh) -> Iterator[PreVisit]:
    def row_handler(url, ts):
        # ok, looks like it's unix epoch
        # https://stackoverflow.com/a/19430099/706389
        dt = datetime.fromtimestamp(int(ts) / 1_000_000, pytz.utc)
        url = unquote(url) # firefox urls are all quoted
        return PreVisit(
            url=url,
            dt=dt,
            locator=Loc.file(histfile),
        )
    yield from browser_extract(
        histfile=histfile,
        cols=cols,
        row_handler=row_handler,
    )

def firefox_phone(histfile: PathIsh) -> Iterator[PreVisit]:
    yield from _firefox(cols=('url', 'date'), histfile=histfile)

def firefox(histfile: PathIsh) -> Iterator[PreVisit]:
    yield from _firefox(cols=('url', 'visit_date'), histfile=histfile)


# should be utc? https://stackoverflow.com/a/26226771/706389
# yep, tested it and looks like utc
def chrome_time_to_utc(chrome_time: int) -> datetime:
    epoch = (chrome_time / 1_000_000) - 11644473600
    return datetime.fromtimestamp(epoch, pytz.utc)


# TODO could use sqlite3 module I guess... but it's quick enough to extract as it is
def chrome(histfile: PathIsh) -> Iterator[PreVisit]:
    def row_handler(url, ts, durs):
        dt = chrome_time_to_utc(int(ts))
        url = unquote(url) # chrome urls are all quoted # TODO not sure if we want it here?
        dd = int(durs)
        dur: Optional[Second]
        if dd == 0:
            dur = None
        else:
            dur = dd // 1_000_000
        return PreVisit(
            url=url,
            dt=dt,
            locator=Loc.file(histfile),
            duration=dur,
        )

    yield from browser_extract(
        histfile=histfile,
        cols=('url', 'visit_time', 'visit_duration'),
        row_handler=row_handler,
    )



import pytest # type: ignore

from promnesia.cannon import canonify, CanonifyException

# TODO should actually understand 'sequences'?
# e.g.
# https://www.scottaaronson.com/blog/?p=3167#comment-1731882 is kinda hierarchy of scottaaronson.com, post 3167 and comment 1731882
# but when working with it from server, would be easier to just do multiple queries I guess..
# https://www.scottaaronson.com/blog/?p=3167 is kind ahierarchy of scottaaronson.com ;


param = pytest.mark.parametrize

# TODO assume spaces are not meaninfgul??
# then could align URLs etc?

@param('url,expected', [(
    'https://www.youtube.com/watch?v=1NHbPN9pNPM&index=63&list=WL&t=491s',
    # TODO not so sure about &t, it's sort of useful
    'youtube.com/watch?list=WL&v=1NHbPN9pNPM'
), (
    'youtube.com/watch?v=wHrCkyoe72U&feature=share&time_continue=6',
    'youtube.com/watch?v=wHrCkyoe72U'
), (
    'youtube.com/embed/nyc6RJEEe0U?feature=oembed',
    'youtube.com/watch?v=nyc6RJEEe0U'
), (
    'https://youtu.be/iCvmsMzlF7o?list=WL',
    # TODO hmm. ordering?
    'youtube.com/watch?list=WL&v=iCvmsMzlF7o'
),
    # TODO can even be like that or contain timestamp (&t=)
    # TODO warn if param already present? shouldn't happen..

    # TODO could be interesting to do automatic rule extraction by querying one represnetative and then extracting canonical

    # TODO national domains don't matter for youtube

    # [*, 'youtube', ANY_DOMAIN] / 'embed' -> 'youtube.com/watch'
    # TODO use regex backrefs?
    #
(
    'm.youtube.com/watch?v=Zn6gV2sdl38',
    'youtube.com/watch?v=Zn6gV2sdl38'
),

    # ( "https//youtube.com/playlist?list=PLeOfc0M-50LmJtZwyOfw6aVopmIbU1t7t"
    # , "youtube.com/playlist?list=PLeOfc0M-50LmJtZwyOfw6aVopmIbU1t7t"
    # ),
    # TODO perhaps it should result in video link + sibling link?
    # when exploring other people's playlists this could be quite useful?

    # ( "https://www.youtube.com/watch?v=1NHbPN9pNPM&index=63&list=WL&t=491s"
    # , "youtube.com/watch?v=1NHbPN9pNPM&list=WL" # TODO not so sure about &t, it's sort of useful
    # ),
    # TODO
    # youtube.com/user/magauchsein/playlists?sort=dd&view=50&shelf_id=14
    # youtube.com/user/TheChemlife/videos?view=0&sort=p&flow=grid
])
def test_youtube(url, expected):
    assert canonify(url) == expected


@param('url,expected', [(
    'https://web.archive.org/web/20090902224414/http://reason.com/news/show/119237.html',
    'reason.com/news/show/119237.html',
)])
def test_archiveorg(url, expected):
    assert canonify(url) == expected


@param('url, expected', [
    ( 'https://www.reddit.com/r/firefox/comments/bbugc5/firefox_bans_free_speech_commenting_plugin/?ref=readnext'
    , 'reddit.com/r/firefox/comments/bbugc5/firefox_bans_free_speech_commenting_plugin',
    ),

    ( 'https://www.reddit.com/r/selfhosted/comments/8j8mo3/what_are_you_self_hosting/dz19gh9/?utm_content=permalink&utm_medium=user&utm_source=reddit&utm_name=u_karlicoss'
    , 'reddit.com/r/selfhosted/comments/8j8mo3/what_are_you_self_hosting/dz19gh9',
    )
    # TODO hmm. parent relationship can just rely on urls for reddit
    # just need to support it in server I suppose

    # TODO search queries?
    # https://www.reddit.com/search?q=AutoValue

    # TODO def need better markdown handling
    # https://reddit.com/r/intj/comments/cmof04/me_irl/ew4a3dw/][    Me_irl]
    # reddit.com/r/intj/comments/cmof04/me_irl/ew4a3dw/%5D%5BMe_irl%5D



])
def test_reddit(url, expected):
    assert canonify(url) == expected

@pytest.mark.parametrize("url,expected", [
    # TODO ?? 'https://groups.google.com/a/list.hypothes.is/forum/#!topic/dev/kcmS7H8ssis',
    #
    # TODO FIXME fragment handling
    # ( "https://www.scottaaronson.com/blog/?p=3167#comment-1731882"
    # , "scottaaronson.com/blog/?p=3167#comment-1731882"
    # ),


    # TODO FIXME fragment handling
    # ( "https://en.wikipedia.org/wiki/tendon#cite_note-14"
    # , "en.wikipedia.org/wiki/tendon#cite_note-14"
    # ),

    # TODO FIXME fragment handling
    # ( "https://physicstravelguide.com/experiments/aharonov-bohm#tab__concrete"
    # , "physicstravelguide.com/experiments/aharonov-bohm#tab__concrete"
    # ),

    ( "https://github.com/search?o=asc&q=track&s=stars&type=Repositories"
    , "github.com/search?q=track"
    ),
    ( "https://80000hours.org/career-decision/article/?utm_source=The+EA+Newsletter&utm_campaign=04ca3c2244-EMAIL_CAMPAIGN_2019_04_03_04_26&utm_medium=email&utm_term=0_51c1df13ac-04ca3c2244-318697649"
    , "80000hours.org/career-decision/article"
    ),
    ( "https://www.facebook.com/photo.php?fbid=24147689823424326&set=pcb.2414778905423667&type=3&theater"
    , "facebook.com/photo.php?fbid=24147689823424326"
    ),
    ( "https://play.google.com/store/apps/details?id=com.faultexception.reader&hl=en"
    , "play.google.com/store/apps/details?id=com.faultexception.reader"
    ),
    # TODO it also got &p= parameter, which refers to page... not sure how to handle this
    # news.ycombinator.com/item?id=15451442&p=2
    ( "https://news.ycombinator.com/item?id=12172351"
    , "news.ycombinator.com/item?id=12172351"
    ),
    ( "https://urbandictionary.com/define.php?term=Belgian%20Whistle"
    , "urbandictionary.com/define.php?term=Belgian%20Whistle"
    ),
    ( "https://en.wikipedia.org/wiki/Dinic%27s_algorithm"
    , "en.wikipedia.org/wiki/Dinic%27s_algorithm"
    ),

    ( "zoopla.co.uk/to-rent/details/42756337#D0zlBWeD4X85odsR.97"
    , "zoopla.co.uk/to-rent/details/42756337"
    ),

    ( "withouthspec.co.uk/rooms/16867952?guests=2&adults=2&location=Berlin%2C+Germany&check_in=2017-08-16&check_out=2017-08-20"
    , "withouthspec.co.uk/rooms/16867952"
    ),

    ( "amp.theguardian.com/technology/2017/oct/09/mark-zuckerberg-facebook-puerto-rico-virtual-reality"
    , "theguardian.com/technology/2017/oct/09/mark-zuckerberg-facebook-puerto-rico-virtual-reality",
    ),

    ( "https://answers.yahoo.com/question/index?qid=20071101131442AAk9bGp"
    , "answers.yahoo.com/question/index?qid=20071101131442AAk9bGp"
    ),
    ( "flowingdata.com/2010/12/14/10-best-data-visualization-projects-of-the-year-%e2%80%93-2010"
    , "flowingdata.com/2010/12/14/10-best-data-visualization-projects-of-the-year-%E2%80%93-2010"
    ),
    ( "flowingdata.com/2010/12/14/10-best-data-visualization-projects-of-the-year-–-2010"
    , "flowingdata.com/2010/12/14/10-best-data-visualization-projects-of-the-year-%E2%80%93-2010"
    ),

    ( "https://spoonuniversity.com/lifestyle/marmite-ways-to-eat-it&usg=AFQjCNH4s1SOEjlpENlfPV5nuvADZpSdow"
    , "spoonuniversity.com/lifestyle/marmite-ways-to-eat-it"
    ),

    ( 'https://google.co.uk/amp/s/amp.reddit.com/r/androidapps/comments/757e2t/swiftkey_or_gboard'
    , 'reddit.com/r/androidapps/comments/757e2t/swiftkey_or_gboard'
    ),

    # should sort query params
    ( 'https://www.youtube.com/watch?v=hvoQiF0kBI8&list=WL&index=2'
    , 'youtube.com/watch?list=WL&v=hvoQiF0kBI8',
    ),
    ( 'https://www.youtube.com/watch?list=WL&v=hvoQiF0kBI8&index=2'
    , 'youtube.com/watch?list=WL&v=hvoQiF0kBI8',
    ),

    # TODO def need to allow the _user_ to define the rules.
    # no way I can predict everything
    # basically, allow *interactively* select
    # also allow introspection, which rule matched?
    ( 'https://bbs.archlinux.org/viewtopic.php?id=212740'
    , 'bbs.archlinux.org/viewtopic.php?id=212740',
    ),

    ( 'https://ubuntuforums.org/showthread.php?t=1403470&s=0dd67bdb12559c22e73a220752db50c7&p=8806195#post8806195'
    , 'ubuntuforums.org/showthread.php?t=1403470',
    ),

    # ( "gwern.net/DNB+FAQ"
    # , "TODO" # ???
    # ),

    # TODO shit. is that normal??? perhaps need to manually move fragment?
    # SplitResult(scheme='https', netloc='unix.stackexchange.com', path='/questions/171603/convert-file-contents-to-lower-case/171708', query='', fragment='171708&usg=AFQjCNEFCGqCAa4P4Zlu2x11bThJispNxQ')
    # ( "https://unix.stackexchange.com/questions/171603/convert-file-contents-to-lower-case/171708#171708&usg=AFQjCNEFCGqCAa4P4Zlu2x11bThJispNxQ"
    # , "unix.stackexchange.com/questions/171603/convert-file-contents-to-lower-case/171708#171708"
    # )
])
def test(url, expected):
    assert canonify(url) == expected
    # TODO github queries
# github.com/search?l=Python&q=reddit+backup
# github.com/search?p=3&q=ipynb+language%3AHaskell
# github.com/search?q=kobo+ExtraData
# github.com/search?q=what-universal-human-experiences-are-you-missing-without-realizing-it

    # TODO git+https://github.com/expectocode/telegram-export@master
    # TODO  again, for that actually sequence would be good...

    # TODO "https://twitter.com/search?q=pinboard search&src=typd"

    # TODO https://www.zalando-lounge.ch/#/
    # TODO m.facebook.com
    # TODO         [R('^(youtube|urbandictionary|tesco|scottaaronson|answers.yahoo.com|code.google.com)') , None],



    # TODO
# amazon.co.uk/gp/offer-listing/B00525XKL4/ref=dp_olp_new
# amazon.co.uk/gp/offer-listing/B00525XKL4/ref=olp_twister_child

    # TODO
    # en.wikipedia.org/wiki/S&P_500_Index


    # TODO
    # google.co.uk/maps/place/Hackney+Bureau/@51.5293789,-0.0527919,16.88z/data=!bla-bla!-bla


    # TODO
    # perhaps, disable utf8 everywhere?
    # github.com/search?utf8=%E2%9C%93&q=%22My+Clippings.txt%22

    # TODO FIXME fragment handling
    # ( "https://www.scottaaronson.com/blog/?p=3167#comment-1731882"
    # , "scottaaronson.com/blog/?p=3167#comment-1731882"
    # ),

@pytest.mark.parametrize("urls", [
    {
        "launchpad.net/ubuntu/%2Bsource/okular",
        "launchpad.net/ubuntu/+source/okular",
    },
    {
        "flowingdata.com/2010/12/14/10-best-data-visualization-projects-of-the-year-–-2010",
        "flowingdata.com/2010/12/14/10-best-data-visualization-projects-of-the-year-%e2%80%93-2010",
        "https://flowingdata.com/2010/12/14/10-best-data-visualization-projects-of-the-year-%e2%80%93-2010/&usg=AFQjCNEZsEGz9rqpWqlFXR5Tc7pkCKY5sQ",
    },
])
def test_same_norm(urls):
    urls = list(sorted(urls))
    u0 = urls[0]
    c0 = canonify(u0)
    for u in urls[1:]:
        c = canonify(u)
        assert c0 == c, f'Expected {u0} and {u} to be same canonically; got {c0} and {c} instead'

def test_error():
    # canonify('  +74Zo535, fewfwf@gmail.com') # -- apparently was patched in some python3.7 versions
    with pytest.raises(CanonifyException):
        # borrowed from https://bugs.mageia.org/show_bug.cgi?id=24640#c7
        canonify('https://example.com\uFF03@bing.com')



#!/usr/bin/env python3
"""
Experimental libarary for normalising domain names and urls to achieve 'canonical' urls for content.
E.g. 
 https://mobile.twitter.com/demarionunn/status/928409560548769792
and
 https://twitter.com/demarionunn/status/928409560548769792
are same content, but you can't tell that by URL equality. Even canonical urls are different!

Also some experiments to establish 'URL hierarchy'.
"""
# TODO eh?? they fixed mobile.twitter.com?

import re
import typing
from typing import Iterable, NamedTuple, Set, Optional, List, Sequence, Union, Tuple, Dict

import urllib.parse
from urllib.parse import urlsplit, parse_qsl, urlunsplit, parse_qs, urlencode, SplitResult


# this has some benchmark, but quite a few librarires seem unmaintained, sadly
# I guess i'll stick to default for now, until it's a critical bottleneck
# https://github.com/commonsearch/urlparse4
# rom urllib.parse import urlparse

# TODO perhaps archive.org contributes to both?

def try_cutl(prefix, s):
    if s.startswith(prefix):
        return s[len(prefix):]
    else:
        return s

def try_cutr(suffix, s):
    if s.endswith(suffix):
        return s[:-len(suffix)]
    else:
        return s

dom_subst = [
    ('m.youtube.'     , 'youtube.'),
    ('studio.youtube.', 'youtube.'),

    ('mobile.twitter.', 'twitter.'),
    ('m.twitter.'     , 'twitter.'),

    ('m.reddit.'      , 'reddit.'),
    ('old.reddit.'    , 'reddit.'),
    ('i.reddit.'      , 'reddit.'),
    ('pay.reddit.'    , 'reddit.'),
    ('np.reddit.'     , 'reddit.'),

    ('m.facebook.'    , 'facebook.'),
]

def canonify_domain(dom: str):
    # TODO perhaps not necessary now that I'm checking suffixes??
    for st in ('www.', 'amp.'):
        dom = try_cutl(st, dom)

    for start, repl in dom_subst:
        if dom.startswith(start):
            dom = repl + dom[len(start):]
            break

    return dom



default_qremove = {
    'utm_source',
    'utm_campaign',
    'utm_content',
    'utm_medium',
    'utm_term',
    'utm_umg_et',

    # https://moz.com/blog/decoding-googles-referral-string-or-how-i-survived-secure-search
    # some google referral
    'usg',

    # google language??
    'hl',
    'vl',

    # e.g. on github
    'utf8',
}

default_qkeep = {
    # ok, various BBS have it, hackernews has it etc?
    # hopefully it's a reasonable one to keep..
    'id',
}

# TODO perhaps, decide if fragment is meaningful (e.g. wiki) or random sequence of letters?
class Spec(NamedTuple):
    qkeep  : Optional[Set[str]] = None
    qremove: Optional[Set[str]] = None
    fkeep  : bool = False

    def keep_query(self, q: str) -> bool:
        qkeep   = default_qkeep  .union(self.qkeep or {})
        qremove = default_qremove.union(self.qremove or {})
        # I suppose 'remove' is only useful for logging. we remove by default anyway

        keep = False
        remove = False
        if qkeep is not None and q in qkeep:
            keep = True
        # pylint: disable=unsupported-membership-test
        if q in qremove:
            remove = True
        if keep and remove:
            # keep wins?
            return True # TODO need a warning
        if keep:
            return True
        if remove:
            return False

        # by default drop all
        # it's a better default, since if it's *too* unified, the user would notice it. but not vice versa!
        return False

    @classmethod
    def make(cls, **kwargs):
        return cls(**kwargs)

S = Spec.make

# TODO perhaps these can be machine learnt from large set of urls?
specs = {
    'youtube.com': S(
        # TODO search_query?
        qkeep={
            'list', # TODO hmm. list is kinda important for playlist urls...
            'v',
        }, # TODO frozenset?
        qremove={
            't', # TODO not so sure about it


            'time_continue', 'index', 'feature', 'lc', 'app', 'start_radio', 'pbjreload', 'annotation_id',
            'flow', 'sort', 'view',
            'enablejsapi', 'wmode', 'html5', 'autoplay', 'ar',

            'gl', # gl=GB??
            'sub_confirmation', 'shelf_id', 'disable_polymer', 'spfreload', 'src_vid', 'origin', 'rel', 'shuffle',
            'nohtml5',
            'showinfo', 'ab_channel', 'start', 'ebc', 'ref', 'view_as', 'fr', 'redirect_to_creator',
            'sp', # TODO ??
            'noapp', 'client', 'sa', 'ob', 'fbclid', 'noredirect', 'zg_or', 'ved',
        } # TODO not so sure about t
    ),
    # TODO shit. for playlist don't need to remove 'list'...


    'github.com': S(
        qkeep={'q'},
        qremove={'o', 's', 'type', 'tab', 'code', 'privacy', 'fork'},
    ),
    'facebook.com': S(
        qkeep={'fbid', 'story_fbid'},
        qremove={
            'set', 'type', 'fref', 'locale2', '__tn__', 'notif_t', 'ref', 'notif_id', 'hc_ref', 'acontext', 'multi_permalinks', 'no_hist', 'next', 'bucket_id',
            'eid',
            'tab', 'active_tab',

            'source', 'tsid', 'refsrc', 'pnref', 'rc', '_rdr', 'src', 'hc_location', 'section', 'permPage', 'soft', 'pn_ref', 'action',
            'ti', 'aref', 'event_time_id', 'action_history', 'filter', 'ref_notif_type', 'has_source', 'source_newsfeed_story_type',
            'ref_notif_type',
        },
    ),
    'physicstravelguide.com': S(fkeep=True), # TODO instead, pass fkeep marker object for shorter spec?
    'wikipedia.org': S(fkeep=True),
    'scottaaronson.com'  : S(qkeep={'p'}, fkeep=True),
    'urbandictionary.com': S(qkeep={'term'}),
    'ycombinator.com'    : S(qkeep={'id'}),
    'play.google.com'    : S(qkeep={'id'}),
    'answers.yahoo.com'  : S(qkeep={'qid'}),
    'ubuntuforums.org'   : S(qkeep={'t'}),
}

_def_spec = S()
# TODO use cache?
def get_spec(dom: str) -> Spec:
    # ugh. a bit ugly way of getting stuff without subdomain...
    parts = dom.split('.')
    cur = None
    for p in reversed(parts):
        if cur is None:
            cur = p
        else:
            cur = p + '.' + cur
        sp = specs.get(cur)
        if sp is not None:
            return sp
    return _def_spec

class CanonifyException(Exception):
    pass

# TODO not so sure if it's better to quote or not?
quote_via   = urllib.parse.quote
unquote_via = urllib.parse.unquote


def _quote_path(path: str) -> str:
    parts = path.split('/')
    nparts = []
    for p in parts:
        # TODO maybe re.match?
        if '%' in p or '+' in p: # some urls are partially encoded... perhaps canonify needs hints indicating if url needs normalising or not
            p = unquote_via(p)
        # TODO safe argumnet?
        nparts.append(quote_via(p))
    return '/'.join(nparts)


# TODO wtf is it doing???
def _prenormalise(url: str) -> str:
    # meh..
    url = re.sub(r'google\..*/amp/s/', '', url)

    if '?' not in url:
        # sometimes urls have not ? but do have query parameters starting with & for some reason; urlsplit chokes over it
        # e.g. in google takeout
        # not sure how safe it in general...
        first_q = url.find('&')
        if first_q != -1:
            return url[:first_q] + '?' + url[first_q + 1:]
    return url


def transform_split(split: SplitResult):
    netloc = canonify_domain(split.netloc)

    path     = split.path
    qparts   = parse_qsl(split.query)

    fragment = split.fragment

    ID   = r'(?P<id>[^/]+)'
    REST = r'(?P<rest>.*)'

    Left = Union[str, Sequence[str]]
    Right = Tuple[str, str, str]
    # the idea is that we can unify certain URLs here and map them to the 'canonical' one
    # this is a dict only for grouping but should be a list really.. todo
    rules: Dict[Left, Right] = {
        # TODO m. handling might be quite common
        # f'm.youtube.com/{REST}': ('youtube.com', '{rest}'),
        (
            f'youtu.be/{ID}',
            f'youtube.com/embed/{ID}',
        ) : ('youtube.com', '/watch', 'v={id}'),
        # TODO wonder if there is a better candidate for canonical video link?
        # {DOMAIN} pattern? implicit?
        (
            'twitter.com/home',
            'twitter.com/explore',
        ) : ('twitter.com', '', ''),
    }

    def iter_rules():
        for fr, to in rules.items():
            if isinstance(fr, str):
                fr = (fr, )
            for f in fr:
                yield f, to

    for fr, to in iter_rules():
        # TODO precache by domain?
        dom, rest = fr.split('/', maxsplit=1)
        if dom != netloc:
            continue

        rest = '/' + rest  # path seems to always start with /
        m = re.fullmatch(rest, path)
        if m is None:
            continue
        gd = m.groupdict()
        if len(to) == 2:
            to = to + ('', )

        (netloc, path, qq) = [t.format(**gd) for t in to]
        qparts.extend(parse_qsl(qq)) # TODO hacky..
        break


    return netloc, path, qparts, fragment



def myunsplit(domain: str, path: str, query: str, fragment: str) -> str:
    uns = urlunsplit((
        '', # dummy protocol
        domain,
        path,
        query,
        fragment,
    ))
    uns = try_cutl('//', uns)  # // due to dummy protocol
    return uns



#
# r'web.archive.org/web/\d+/
# def canonify_simple(url: str) -> Optional[str]:
#     '''
#     Experiment with simply using regexes for normalising, as close to regular python as it gets
#     '''
#     # - using named groups for comments?

#     # TODO to be fair, archive.org is a 'very' special case..
#     regexes = [
#         r'web.archive.org/web/(?<timestamp>\d+)/' # TODO what about 'rest'?
#     ]
#     for re in regexes:

def handle_archive_org(url: str) -> Optional[str]:
    are = r'web.archive.org/web/(?P<timestamp>\d+)/(?P<rest>.*)'
    m = re.fullmatch(are, url)
    if m is None:
        return None
    else:
        return m.group('rest')


# TODO ok, I suppose even though we can't distinguish + and space, likelihood of them overlapping in normalised url is so low, that it doesn't matter much
# TODO actually, might be easier for most special charaters
def canonify(url: str) -> str:
    # TODO check for invalid charaters?
    url = _prenormalise(url)

    try:
        # TODO didn't really get difference from urlparse
        parts = urlsplit(url)
    except Exception as e:
        # TODO not sure about the exception...
        # TODO return it instead of raising
        raise CanonifyException(url) from e

    # TODO move to prenormalise?
    if parts.scheme == '':
        # if scheme is missing it doesn't parse netloc properly...
        try:
            parts = urlsplit('http://' + url)
        except Exception as e:
            raise CanonifyException(url) from e

    # meh. figure out what's up with transform_split
    no_protocol = parts.netloc + parts.path + parts.query + parts.fragment

    res = handle_archive_org(no_protocol)
    if res is not None:
        assert len(res) < len(no_protocol) # just a paranoia to avoid infinite recursion...
        return canonify(res)

    domain, path, qq, _frag = transform_split(parts)

    spec = get_spec(domain)

    # TODO FIXME turn this logic back on?
    # frag = parts.fragment if spec.fkeep else ''
    frag = ''

    qq = [(k, v) for k, v in qq if spec.keep_query(k)]
    qq = list(sorted(qq)) # order matters in theory, but definitely not by default
    # TODO still not sure what we should do..
    # quote_plus replaces %20 with +, not sure if we want it...
    query = urlencode(qq, quote_via=quote_via)

    path = _quote_path(path)

    uns = myunsplit(domain, path, query, frag)
    uns = try_cutr('/', uns) # not sure if there is a better way
    return uns


 # TODO wonder if lisp could be convenient for this. lol
TW_PATTERNS = [
    {
        'U': r'[\w-]+',
        'S': r'\d+',
        'L': r'[\w-]+',
    },

    r'twitter.com/U/status/S',
    r'twitter.com/U/status/S/retweets',
    r'twitter.com/statuses/S',
    r'twitter.com/U',
    r'twitter.com/U/likes',
    r'twitter.com/U/status/S/.*',
    r'twitter.com/U/statuses/S',
    r'twitter.com/i/notifications',
    r'twitter.com/U/with_replies',
    r'twitter.com/(settings|account)/.*',
    r'twitter.com/U/lists',
    r'twitter.com/U/lists/L(/.*)?',
    r'twitter.com/i/web/status/S',
    r'twitter.com',
    r'twitter.com/home',
    r'tweetdeck.twitter.com',
    r'twitter.com/U/(media|photo|followers|following)',
    # r'mobile.twitter.com/.*',

    r'twitter.com/compose/tweet',
    r'twitter.com/hashtag/\w+',
    r'twitter.com/i/(moments|offline|timeline|u|bookmarks|display|redirect)',
    r'twitter.com/intent/tweet.*',
    r'twitter.com/lists/(create|add_member)',
    r'twitter.com/search-home',
    r'twitter.com/notifications/\d+',

    r'twitter.com/i/events/\d+',

    r'(dev|api|analytics|developer|help|support|blog|anywhere|careers|pic).twitter.com/.*',
]

RD_PATTERNS = [
    {
        'U'  : r'[\w-]+',
        'S'  : r'\w+',
        'PID': r'\w+',
        'PT' : r'\w+',
        'CID': r'\w+',
    },
    r'reddit.com/(user|u)/U',
    r'reddit.com/user/U/(comments|saved|posts)',
    r'reddit.com/r/S(/(top|new|hot|about|search|submit|rising))?',
    r'reddit.com/r/S/comments/PID/PT',
    r'reddit.com/r/S/duplicates/PID/PT',
    r'reddit.com/r/S/comments/PID/PT/CID',
    r'reddit.com/r/S/wiki(/.*)?',

    r'reddit.com/comments/PID',
    r'reddit.com/(subreddits|submit|search|top|original|premium|gilded|gold|gilding)',
    r'reddit.com/subreddits/mine',
    r'reddit.com',
    r'reddit.com/(settings|prefs)(/.*)?',
    r'reddit.com/message/(unread|inbox|messages|compose|sent|newsletter)',
    r'reddit.com/me/.*',
    r'reddit.com/login',

    r'ssl.reddit.com/api/v1/authorize',
    r'reddit.com/dev/api',
    r'reddit.com/api/v1/authorize',
    r'reddit.com/domain/.*',
]

GH_PATTERNS = [
    {
        'U': r'[\w-]+',
        'R': r'[\w\.-]+',
        'b': r'[\w-]+',
        'X': r'(/.*)?',
        'C': r'\w+',
        'G': r'\w+',
    },
    r'github.com/U',
    r'github.com/U/R(/(watchers|settings|actions|branches))?',
    r'github.com/U/R/wiki/.*',
    r'github.com/U/R/issuesX',
    r'github.com/U/R/issues\?q.*',
    r'github.com/U/R/networkX',
    r'github.com/U/R/releasesX',
    r'github.com/U/R/blobX',
    r'github.com/U/R/treeX',
    r'github.com/U/R/pullX',
    r'github.com/U/R/pulls',
    r'github.com/U/R/wiki',
    r'github.com/U/R/commits/B',
    r'github.com/U/R/commit/C',
    r'github.com/U/R/search\?.*',
    r'github.com/search\?.*',
    r'github.com/search/advanced\?.*',
    r'github.com/loginX',
    r'github.com/settings/.*',
    r'github.com/\?q.*',

    r'github.com/_google_extract.*', # TODO wtf?
    r'github.communityX',
    r'github.com/dashboard',
    r'help.github.comX',
    r'gist.github.com/UX',
    r'gist.github.com/G',
    r'developer.github.com/.*',

    # TODO FIXME no canonical here
    # https://gist.github.com/dneto/2258454
    # same as https://gist.github.com/2258454
]

YT_PATTERNS = [
    {
        'V': r'[\w-]+',
        'L': r'[\w-]+',
        'U': r'[\w-]+',
        'C': r'[\w-]+',
    },
    r'youtube.com/watch\?v=V',
    r'youtube.com/watch\?list=L&v=V',
    r'youtube.com/watch\?list=L',
    r'youtube.com/playlist\?list=L',
    r'youtube.com/user/U(/(videos|playlists|feautred|channels|featured))?',
    r'youtube.com/(channel|c)/C(/(videos|playlists))?',

    r'accounts.youtube.com/.*',
    r'youtube.com/signin\?.*',
    r'youtube.com/redirect\?.*',
    r'youtube.com/results\?(search_query|q)=.*',
    r'youtube.com/feed/(subscriptions|library|trending|history)',
    r'youtube.com',
    r'youtube.com/(post_login|upload)',
]

SOP = r'(^|\w+\.)stackoverflow.com'

SO_PATTERNS = [
    {
        # TODO just replace with identifier? should be quite unambiguous
        'QI': r'\d+',
        'QT': r'[\w-]+',
        'A' : r'\d+',
        'UI': r'\d+',
        'U' : r'[\w-]+',
    },
    SOP + r'/questions/QI/QT',
    SOP + r'/questions/QI/QT/A',
    SOP + r'/q/QI',
    SOP + r'/q/QI/A',
    SOP + r'/a/QI',
    SOP + r'/a/QI/A',
    SOP + r'/search',
    SOP + r'/users/UI',
    SOP + r'/users/UI/U',
    SOP,
]

WKP = r'(^|.+\.)wikipedia.org'

WK_PATTERNS = [
    {
        'AN': r'[\w%.-]+',
    },
    WKP + '/wiki/AN',
    WKP,
]

FB_PATTERNS = [
    {
        'F': 'facebook.com',
        'U': r'[\w\.-]+',
        'P': r'\d+',
        'I': r'\d+',
        'CI': r'\d+',
    },
    r'F',
    r'F/U',
    r'F/U/(posts|videos)/P',
    r'F/U/posts/P\?comment_id=CI',
    r'F/photo.php\?fbid=I',
    r'F/photo.php\?fbid=I&id=I',
    r'F/profile.php\?fbid=I',
    r'F/profile.php\?id=I',
    r'F/groups/U',
    r'F/search/.*',
    r'F/events/I',
    r'F/events/I/permalink/I',
    r'F/events/I/I',
    r'F/U/photos/pcb.I/I',

    r'F/pages/U/P',
    r'F/stories/I',
    r'F/notes/U/P',
]

PATTERNS = {
    'twitter'   : TW_PATTERNS,
    'reddit'    : RD_PATTERNS,
    'github.com': GH_PATTERNS,
    'youtube'   : YT_PATTERNS,
    'stackoverflow': SO_PATTERNS,
    'facebook'  : FB_PATTERNS,
    'wikipedia' : WK_PATTERNS,
}


def get_patterns():  # pragma: no cover
    def repl(p, dct):
        for k, v in dct.items():
            p = p.replace(k, v)
        # TODO FIXME unreplaced upper?
        return p

    def handle(stuff):
        pats = []
        repls = []
        for x in stuff:
            if isinstance(x, dict):
                repls.append(x)
            else:
                pats.append(x)
        if len(repls) == 0:
            repls.append({})
        [rdict] = repls
        for p in pats:
            yield repl(p, rdict)
    return {k: list(handle(v)) for k, v in PATTERNS.items()}


def domains(it): # pragma: no cover
    from collections import Counter
    c: typing.Counter[str] = Counter()
    for line in it:
        url = line.strip()
        try:
            nurl = canonify(url)
        except CanonifyException as e:
            print(f"ERROR while normalising! {nurl} {e}")
            c['ERROR'] += 1
            continue
        else:
            udom = nurl[:nurl.find('/')]
            c[udom] += 1
    from pprint import pprint
    pprint(c.most_common(20))


def groups(it, args): # pragma: no cover
    all_pats = get_patterns()

    from collections import Counter
    c: typing.Counter[Optional[str]] = Counter()
    unmatched: List[str] = []

    def dump():
        print(c)
        print('   ' + str(unmatched[-10:]))

    def reg(url, pat):
        c[pat] += 1
        if pat is None:
            unmatched.append(nurl)


    for i, line in enumerate(it):
        if i % 10000 == 0:
            pass
            # dump()
        url = line.strip()
        try:
            nurl = canonify(url)
        except CanonifyException as e:
            print(f"ERROR while normalising! {nurl} {e}")
            continue
        udom = nurl[:nurl.find('/')]
        usplit = udom.split('.')
        patterns = None
        for dom, pats in all_pats.items():
            dsplit = dom.split('.')
            if '$'.join(dsplit) in '$'.join(usplit): # meh
                patterns = pats
                break
        else:
            # TODO should just be ignored?
            # reg(nurl, None)
            continue

        pat = None
        for p in patterns:
            m = re.fullmatch(p, nurl)
            if m is None:
                continue
            pat = p
            break
        reg(nurl, pat)
    for u in sorted(unmatched):
        print(u)
    dump()
    nones = c[None]
    # TODO print link examples alongside?
    print(f"Unmatched: {nones / sum(c.values()) * 100:.1f}%")
    uc = Counter([u.split('/')[:2][-1] for u in unmatched]).most_common(10)
    from pprint import pprint
    pprint(uc)



def display(it, args): # pragma: no cover
    # TODO better name?
    import difflib
    # pylint: disable=import-error
    from termcolor import colored as C # type: ignore
    from sys import stdout

    for line in it:
        line = line.strip()
        if args.human:
            print('---')
            print(line)
        can = canonify(line)
        # TODO use textual diff?
        sm = difflib.SequenceMatcher(None, line, can)

        org_ = ""
        can_ = ""

        pr = False
        def delete(x):
            nonlocal pr
            if x in (
                    'https://www.',
                    'http://www.',
                    'http://',
                    'https://',
                    'file://',
                    '/',
            ):
                col = None
            else:
                if len(x) > 0:
                    pr = True
                col = 'red'
            return C(x, color=col)

        for what, ff, tt, ff2, tt2 in sm.get_opcodes():
            if what == 'delete':
                fn = delete
            elif what == 'equal':
                fn = lambda x: C(x, color=None)
            else:
                pr = True
                fn = lambda x: C(x, color='cyan')
            # TODO exclude certain items from comparison?


            org_ += fn(line[ff: tt])
            can_ += fn(can[ff2: tt2])
            cl = max(len(org_), len(can_))
            org_ += ' ' * (cl - len(org_))
            can_ += ' ' * (cl - len(can_))

        if pr:
            stdout.write(f'{org_}\n{can_}\n---\n')


def main(): # pragma: no cover
    import argparse
    p = argparse.ArgumentParser(epilog='''
- sqlite3 promnesia.sqlite 'select distinct orig_url from visits' | cannon.py --domains

- running comparison
  sqlite3 promnesia.sqlite 'select distinct orig_url from visits where norm_url like "%twitter%" order by orig_url' | src/promnesia/cannon.py
''', formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=100) # type: ignore
    )
    p.add_argument('input', nargs='?')
    p.add_argument('--human', action='store_true')
    p.add_argument('--groups', action='store_true')
    p.add_argument('--domains', action='store_true', help='useful for counting number of URLs by domains')
    args = p.parse_args()

    it: Iterable[str]
    if args.input is None:
        import sys
        it = sys.stdin
    else:
        it = [args.input]

    if args.groups:
        groups(it, args)
    if args.domains:
        domains(it)
    else:
        display(it, args)


if __name__ == '__main__':
    main() # pragma: no cover

# TODO hmm, it's actually sort of fingerprinter... so maybe that's what I should call it

# tweet          -P-> user (not always possible to determine)
# reddit comment -P-> reddit post
# reddit post    -P-> subreddit
# YT video       -P-> user, playlist

# wikipedia footnote will have ???
# certain urls would need to treat # as parent relationship (e.g. slatestarcodex?)

# siblings are items that share a parent (could also return which one is shared)

# youtube end domain normalising: very few occurences, so I won't care about them for now


# TODO for debugging, have special mode that only uses query param trimming one by one
# TODO show percents?

# TODO ugh. need coverage...



import logging

import pytz
from itertools import chain
from datetime import datetime
from typing import List, Optional, Iterable
from pathlib import Path
import json

from my.kython.kompress import kexists, kopen

from ..common import Visit, get_logger, PathIsh, Url, Loc
from .. import config


from more_itertools import unique_everseen
from cachew import mtime_hash, cachew


# TODO use CPath? Could encapsulate a path within an archive *or* within a directory
TakeoutPath = Path


# TODO should this be HPI responsibility?
def cacheme(ident: str):
    logger = get_logger()
    # doesn't even need a nontrivial hash function, timestsamp is encoded in name
    def db_pathf(takeout: TakeoutPath) -> Path:
        tpath = Path(str(takeout))
        cname = tpath.name + '_' + ident + '.cache'
        if config.has(): # TODO eh?
            cache_dir = Path(config.get().cache_dir)
        else:
            # TODO hmm. if using relative path, make it relative to /tmp?
            logger.warning('Caching in /tmp')
            cache_dir = Path('/tmp')
        return cache_dir / cname
    return cachew(db_pathf, cls=Visit, logger=logger)


def _read_myactivity_html(takeout: TakeoutPath, kind: str) -> Iterable[Visit]:
    logger = get_logger()
    # TODO glob
    # TODO not sure about windows path separators??
    spath = 'Takeout/My Activity/' + kind
    if not kexists(takeout, spath):
        logger.warning(f"{spath} is not present in {takeout}... skipping")
        return []
    logger.info('processing %s %s', takeout, kind)

    locator = Loc.file(spath)
    from my.google.takeout.html import read_html
    for dt, url, title in read_html(takeout, spath):
        yield Visit(
            url=url,
            dt=dt,
            locator=locator,
            debug=kind,
        )


@cacheme('google_activity')
def read_google_activity(takeout: TakeoutPath) -> Iterable[Visit]:
    return _read_myactivity_html(takeout, 'Chrome/MyActivity.html')

@cacheme('search_activity')
def read_search_activity(takeout: TakeoutPath) -> Iterable[Visit]:
    return _read_myactivity_html(takeout, 'Search/MyActivity.html')


# TODO add this to tests?
@cacheme('browser_activity')
def read_browser_history_json(takeout: TakeoutPath) -> Iterable[Visit]:
    # not sure if this deserves moving to HPI? it's pretty trivial for now
    logger = get_logger()
    spath = 'Takeout/Chrome/BrowserHistory.json'

    if not kexists(takeout, spath):
        logger.warning(f"{spath} is not present in {takeout}... skipping")
        return
    logger.info('processing %s %s', takeout, spath)

    # TODO couls also add spath?
    locator = Loc.file(takeout)

    j = None
    with kopen(takeout, spath) as fo: # TODO iterative parser?
        j = json.load(fo)

    hist = j['Browser History']
    for item in hist:
        url = item['url']
        time = datetime.utcfromtimestamp(item['time_usec'] / 10 ** 6).replace(tzinfo=pytz.utc)
        # TODO any more interesitng info?
        yield Visit(
            url=url,
            dt=time,
            locator=locator,
            debug='Chrome/BrowserHistory.json',
        )


# TODO make an iterator, insert in db as we go? handle errors gracefully?
def index() -> Iterable[Visit]:
    from my.google.takeout.paths import get_takeouts
    takeouts = list(get_takeouts())
    # TODO if no takeouts, raise?
    # although could raise a warning on top level, when source emitted no takeouts

    # TODO youtube?
    google_activities = [read_google_activity(t)      for t in takeouts]
    search_activities = [read_search_activity(t)      for t in takeouts]
    browser_histories = [read_browser_history_json(t) for t in takeouts]

    key = lambda v: (v.dt, v.url)
    return chain(
        unique_everseen(chain(*google_activities), key=key),
        unique_everseen(chain(*search_activities), key=key),
        unique_everseen(chain(*browser_histories), key=key),
    )
